"""Implementation of ClusterSubspace and related PottsSubspace classes.

The ClusterSubspace class is the workhorse for generating the objects and
information necessary for a cluster expansion.

The PottsSubspace class is an (experimental) class that is similar, but
diverges from the CE mathematic formalism.
"""
# pylint: disable=too-many-lines


import warnings
from copy import deepcopy
from importlib import import_module

import numpy as np
from monty.json import MSONable
from pymatgen.analysis.structure_matcher import (
    OrderDisorderElementComparator,
    StructureMatcher,
)
from pymatgen.core import PeriodicSite, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer, SymmOp
from pymatgen.util.coord import (
    coord_list_mapping_pbc,
    is_coord_subset,
    is_coord_subset_pbc,
    lattice_points_in_supercell,
)

from smol.cofe.space import (
    Orbit,
    Vacancy,
    basis_factory,
    get_allowed_species,
    get_site_spaces,
)
from smol.cofe.space.basis import IndicatorBasis
from smol.cofe.space.constants import SITE_TOL
from smol.correlations import corr_from_occupancy
from smol.exceptions import SYMMETRY_ERROR_MESSAGE, StructureMatchError, SymmetryError

__author__ = "Luis Barroso-Luque, William Davidson Richards"


class ClusterSubspace(MSONable):
    """ClusterSubspace represents a subspace of functions of configuration.

    A :class:`ClusterSubspace` is the main work horse used in constructing a
    cluster expansion. It is necessary to define the terms to be included in a
    cluster expansion. A cluster subspace holds a finite set of orbits that
    contain symmetrically equivalent clusters. The orbits also contain the set
    of orbit basis functions (also known as correlation functions) that
    represent the terms in the cluster expansion. Taken together the set of all
    orbit functions for all orbits included span a subspace of the total
    function space over the configurational space of a given crystal structure
    system.

    The :class:`ClusterSubspace` also has methods to match fitting structures
    and determine site mappings for supercells of different sizes in order to
    compute correlation vectors (i.e. evaluate the orbit functions for a given
    structure).

    You probably want to generate from :code:`ClusterSubspace.from_cutoffs`,
    which will auto-generate the orbits from diameter cutoffs.

    Attributes:
        symops (list of SymmOp):
            Symmetry operations of structure.
        num_corr_functions (int):
            Total number of correlation functions (orbit basis functions)
            included in the subspace.
        num_orbits (int):
            Total number of crystallographic orbits included in the subspace.
            This includes the empty orbit.
        num_clusters (int):
            Total number of clusters in the primitive cell that are included
            in the subspace.
    """

    def __init__(
        self,
        structure,
        expansion_structure,
        symops,
        orbits,
        supercell_matcher=None,
        site_matcher=None,
        **matcher_kwargs,
    ):
        """Initialize a ClusterSubspace.

        You rarely will need to create a ClusterSubspace using the main
        constructor.
        Look at the class method :code:`from_cutoffs` for the "better" way to
        instantiate a ClusterSubspace.

        Args:
            structure (Structure):
                Structure to define the cluster space. Typically the primitive
                cell. Includes all species regardless of partial occupation.
            expansion_structure (Structure):
                Structure including only sites that will be included in the
                Cluster space (i.e. only those with partial occupancy)
            symops (list of Symmop):
                list of Symmops for the given structure.
            orbits (dict): {size: list of Orbits}
                Dictionary with size (number of sites) as keys and list of
                Orbits as values.
            supercell_matcher (StructureMatcher): optional
                StructureMatcher used to find supercell matrices
                relating the prim structure to other structures. If you pass
                this directly you should know how to set the matcher up,
                otherwise matching your relaxed structures can fail, alot.
            site_matcher (StructureMatcher): optional
                StructureMatcher used to find site mappings
                relating the sites of a given structure to an appropriate
                supercell of the prim structure . If you pass this directly you
                should know how to set the matcher up, otherwise matching your
                relaxed structures can fail, alot.
            matcher_kwargs:
                ltol, stol, angle_tol, supercell_size: parameters to pass
                through to the StructureMatchers. Structures that don't match
                to the primitive cell under these tolerances won't be included
                in the expansion. Easiest option for supercell_size is usually
                to use a species that has a constant amount per formula unit.
                See pymatgen documentation of :class:`StructureMatcher` for
                more details.
        """
        # keep as private attributes
        self._structure = structure
        self._exp_structure = expansion_structure

        self.symops = symops  # should we even keep this as an attribute?
        self.num_corr_functions = None  # set automattically when assigning ids
        self.num_orbits = None  # same as above
        self.num_clusters = None  # same as above

        # Test that all the found symmetry operations map back to the input
        # structure otherwise you can get weird subset/superset bugs.
        fcoords = self._structure.frac_coords
        for sym_op in self.symops:
            if not is_coord_subset_pbc(
                sym_op.operate_multi(fcoords), fcoords, SITE_TOL
            ):
                raise SymmetryError(SYMMETRY_ERROR_MESSAGE)

        # This structure matcher is used to determine if a given (supercell)
        # structure matches the prim structure by retrieving the matrix
        # relating them. Only the "get_supercell_matrix" method is used.
        if supercell_matcher is None:
            sc_comparator = OrderDisorderElementComparator()
            self._sc_matcher = StructureMatcher(
                primitive_cell=False,
                attempt_supercell=True,
                allow_subset=True,
                comparator=sc_comparator,
                scale=True,
                **matcher_kwargs,
            )
        else:
            self._sc_matcher = supercell_matcher

        # This structure matcher is used to find the mapping between the sites
        # of a given supercell structure and the sites in the appropriate sized
        # supercell of the prim structure. Only "get_mapping" method is used.
        if site_matcher is None:
            site_comparator = OrderDisorderElementComparator()
            self._site_matcher = StructureMatcher(
                primitive_cell=False,
                attempt_supercell=False,
                allow_subset=True,
                comparator=site_comparator,
                scale=True,
                **matcher_kwargs,
            )
        else:
            self._site_matcher = site_matcher

        self._orbits = orbits
        self._external_terms = []  # List will hold external terms (i.e. Ewald)

        # Dict to cache orbit index mappings, this prevents doing another
        # structure match with the _site_matcher for structures that have
        # already been matched
        self._supercell_orb_inds = {}

        # assign the cluster ids
        self._assign_orbit_ids()

    @classmethod
    def from_cutoffs(
        cls,
        structure,
        cutoffs,
        basis="indicator",
        orthonormal=False,
        use_concentration=False,
        supercell_matcher=None,
        site_matcher=None,
        **matcher_kwargs,
    ):
        """Create a ClusterSubspace from diameter cutoffs.

        Creates a :class:`ClusterSubspace` with orbits of the given size and
        diameter smaller than or equal to the given value. The diameter of an
        orbit is the maximum distance between any two sites of a cluster of
        that orbit.

        This is the best (and the only easy) way to create a
        :class:`ClusterSubspace`.

        Args:
            structure (Structure):
                disordered structure to build a cluster expansion for.
                Typically the primitive cell
            cutoffs (dict):
                dict of {cluster_size: diameter cutoff}. Cutoffs should be
                strictly decreasing. Typically something like {2:5, 3:4}.
                The empty orbit is always included. Singlets are by default
                included, with the exception below.
                To obtain a subspace with only an empty and singlet terms use
                an empty dict {}, or {1: 1}. Adding a cutoff term for point
                terms, i.e. {1: 0} is useful to exclude point terms. Any other
                value for the cutoff will simply be ignored.
            basis (str):
                a string specifying the site basis functions
            orthonormal (bool):
                whether to enforce an orthonormal basis. From the current
                available bases only the indicator basis is not orthogonal out
                of the box
            use_concentration (bool):
                if True, the concentrations in the prim structure sites will be
                used to orthormalize site bases. This gives gives a cluster
                subspace centered about the prim composition.
            supercell_matcher (StructureMatcher): optional
                StructureMatcher used to find supercell matrices
                relating the prim structure to other structures. If you pass
                this directly you should know how to set the matcher up,
                otherwise matching your relaxed structures will fail, alot.
            site_matcher (StructureMatcher): optional
                StructureMatcher used to find site mappings
                relating the sites of a given structure to an appropriate
                supercell of the prim structure . If you pass this directly you
                should know how to set the matcher up, otherwise matching your
                relaxed structures will fail, alot.
            matcher_kwargs:
                ltol, stol, angle_tol, supercell_size: parameters to pass
                through to the StructureMatchers. Structures that don't match
                to the primitive cell under these tolerances won't be included
                in the expansion. Easiest option for supercell_size is usually
                to use a species that has a constant amount per formula unit.

        Returns:
            ClusterSubspace
        """
        # get symmetry operations of prim structure.
        symops = SpacegroupAnalyzer(structure).get_symmetry_operations()
        # get the active sites (partial occupancy) to expand over.
        sites_to_expand = [
            site
            for site in structure
            if site.species.num_atoms < 0.99 or len(site.species) > 1
        ]
        expansion_structure = Structure.from_sites(sites_to_expand)
        # get orbits within given cutoffs
        orbits = cls._gen_orbits_from_cutoffs(
            expansion_structure, cutoffs, symops, basis, orthonormal, use_concentration
        )
        return cls(
            structure=structure,
            expansion_structure=expansion_structure,
            symops=symops,
            orbits=orbits,
            supercell_matcher=supercell_matcher,
            site_matcher=site_matcher,
            **matcher_kwargs,
        )

    @property
    def basis_type(self):
        """Get the type of site basis set used."""
        return self.orbits[0].basis_type

    @property
    def structure(self):
        """Get the underlying primitive structure including inactive sites."""
        return self._structure

    @property
    def expansion_structure(self):
        """Get the primitive expansion structure (excludes inactive sites)."""
        return self._exp_structure

    @property
    def cutoffs(self):
        """Return dict of orbit cluster cutoffs.

        These are "tight" cutoffs, as in the maximum diameter for each cluster
        size, which is <= the input to from_cutoffs.
        """
        return {
            size: max(orbit.base_cluster.diameter for orbit in orbits)
            for size, orbits in self._orbits.items()
            if size != 1
        }

    @property
    def orbits(self):
        """Return a list of all orbits sorted by size."""
        return [orbit for _, orbits in sorted(self._orbits.items()) for orbit in orbits]

    @property
    def orbits_by_size(self):
        """Get dictionary of orbits with key being the orbit size."""
        return self._orbits

    @property
    def orbit_multiplicities(self):
        """Get the crystallographic multiplicities for each orbit."""
        mults = [1] + [orb.multiplicity for orb in self.orbits]
        return np.array(mults)

    @property
    def num_functions_per_orbit(self):
        """Get the number of correlation functions for each orbit.

        The list returned has length equal to the total number of orbits,
        and each entry is the total number of correlation functions
        associated with that orbit.
        """
        return np.array([len(orbit) for orbit in self.orbits])

    @property
    def function_orbit_ids(self):
        """Get Orbit IDs corresponding to each correlation function.

        If the ClusterSubspace includes external terms, these are not included
        in the list since they are not associated with any orbit.
        """
        func_orb_ids = [0]
        for orbit in self.orbits:
            func_orb_ids += len(orbit) * [
                orbit.id,
            ]
        return np.array(func_orb_ids)

    @property
    def function_inds_by_size(self):
        """Get correlation function indices by cluster sizes."""
        return {
            s: list(range(os[0].bit_id, os[-1].bit_id + len(os[-1])))
            for s, os in self._orbits.items()
        }

    @property
    def function_ordering_multiplicities(self):
        """Get array of ordering multiplicity of each correlation function.

        The length of the array returned is the total number of correlation
        functions in the subspace for all orbits. The ordering multiplicity of
        a correlation function is the number of symmetrically equivalent bit
        orderings (function-labeled orbit configurations) that result in the
        product of the same single site functions.
        """
        mults = [1] + [
            mult for orb in self.orbits for mult in orb.bit_combo_multiplicities
        ]
        return np.array(mults)

    @property
    def function_total_multiplicities(self):
        """Get array of total multiplicity of each correlation function.

        The length of the array returned is the total number of correlation
        functions in the subspace for all orbits. The total multiplicity of a
        correlation function is the number of symmetrically equivalent bit
        orderings (or function-labeled orbit configurations) that result in the
        product of the same single site functions times the (crystallographic)
        multiplicity of the orbit.
        """
        return (
            self.orbit_multiplicities[self.function_orbit_ids]
            * self.function_ordering_multiplicities
        )

    @property
    def basis_orthogonal(self):
        """Check if the orbit basis is orthogonal."""
        return all(orb.basis_orthogonal for orb in self.orbits)

    @property
    def basis_orthonormal(self):
        """Check if the orbit basis is orthonormal."""
        return all(orb.basis_orthonormal for orb in self.orbits)

    @property
    def external_terms(self):
        """Get external terms to be fitted together with the correlations.

        External terms are those represented by pair interaction Hamiltonians
        (i.e. Ewald electrostatics).
        """
        return self._external_terms

    def orbit_hierarchy(self, level=1, min_size=1):
        """Get orbit hierarchy by IDs.

        The orbit hierarchy represents in inclusion relationships between orbits and
        their suborbits.

        The empty/constant cluster index 0 is technically a suborbit of all
        orbits, but is not added to the hierarchy entries.

        Args:
            level (int):
            min_size (int):

        Returns:
            list of list: each element of the inner lists is the orbit id for
            all suborbits corresponding to the orbit at the given outer list
            index.
        """
        sub_ids = [
            [
                suborb.id
                for suborb in self.get_sub_orbits(
                    orb.id, level=level, min_size=min_size
                )
            ]
            for orb in self.orbits
        ]

        return [
            [],
        ] + sub_ids

    def function_hierarchy(self, level=1, min_size=2, invert=False):
        """Get the correlation function hierarchy.

        The function hierarchy is the relationship between specific correlation
        functions and "sub" correlation functions (i.e. a correlation function is
        a "sub" correlation factor or included in higher degree correlation function if
        it is a factor of a higher degree correlation function.

        Args:
            level (int):
            min_size (int):
            invert (bool): optional
                Default is invert=False which gives the high to low bit combo
                hierarchy. Invert= True will invert the hierarchy into low to
                high

        Returns:
            list of list: each element of the inner lists is the bit id for
            all correlation functions corresponding to the corr function at the given
            outer list index.
        """
        hierarchy = [
            self.get_sub_function_ids(i, level=level, min_size=min_size)
            for i in range(self.num_corr_functions)
        ]

        if invert:
            hierarchy = invert_mapping(hierarchy)

        return hierarchy

    def orbits_from_cutoffs(self, upper, lower=0):
        """Get orbits with clusters within given diameter cutoffs (inclusive).

        Args:
            upper (float or dict):
                upper diameter for clusters to include. If a single float
                is given then that cutoff is used for all orbit sizes.
                Otherwise a dict can be used to specify the cutoff for the
                orbit cluster sizes,
                i.e. {2: pair_cutoff, 3: triplet_cutoff, ...}
            lower (float): optional
                lower diameter for clusters to include. If a single float
                is given then that cutoff is used for all orbit sizes.
                Otherwise a dict can be used to specify the cutoff for the
                orbit cluster sizes,
                i.e. {2: pair_cutoff, 3: triplet_cutoff, ...}

        Returns:
            list of Orbits
        """
        upper = (
            upper
            if isinstance(upper, dict)
            else {k: upper for k in self._orbits.keys()}
        )
        lower = (
            lower
            if isinstance(lower, dict)
            else {k: lower for k in self._orbits.keys()}
        )
        return [
            orbit
            for size in upper.keys()
            for orbit in self._orbits[size]
            if lower[size] <= orbit.base_cluster.diameter <= upper[size]
        ]

    def function_inds_from_cutoffs(self, upper, lower=0):
        """Get indices of correlation functions by cluster cutoffs.

        Args:
            upper (float or dict):
                upper diameter for clusters to include. If a single float
                is given then that cutoff is used for all orbit sizes.
                Otherwise a dict can be used to specify the cutoff for the
                orbit cluster sizes,
                i.e. {2: cutoff_pairs, 3: cutoff_trips, ...}
            lower (float): optional
                lower diameter for clusters to include. If a single float
                is given then that cutoff is used for all orbit sizes.
                Otherwise a dict can be used to specify the cutoff for the
                orbit cluster sizes,
                i.e. {2: cutoff_pairs, 3: cutoff_trips, ...}

        Returns:
            list: list of correlation function indices for clusters within cutoffs
        """
        orbits = self.orbits_from_cutoffs(upper, lower)
        inds = []
        for orbit in orbits:
            inds += list(range(orbit.bit_id, orbit.bit_id + len(orbit)))
        return np.array(inds)

    def add_external_term(self, term):
        """Add an external term to the ClusterSubspace.

        Adds an external term (e.g. an Ewald term) to the cluster expansion
        terms. External term classes must be MSONable and implement a method
        to obtain a "correlation". See smol.cofe.extern for notebooks.

        Args:
            term (ExternalTerm):
                An instance of an external term. Currently only EwaldTerm is
                implemented.
        """
        for added_term in self.external_terms:
            if isinstance(term, type(added_term)):
                raise ValueError(f"This ClusterSubspaces already has an {type(term)}.")
        self._external_terms.append(term)

    @staticmethod
    def num_prims_from_matrix(scmatrix):
        """Get number of prim structures in a supercell for a given matrix."""
        return int(round(np.abs(np.linalg.det(scmatrix))))

    def corr_from_structure(
        self, structure, normalized=True, scmatrix=None, site_mapping=None
    ):
        """Get correlation vector for structure.

        Returns the correlation vector for a given structure. To do this, the
        correct supercell matrix of the prim needs to be found to then
        determine the mappings between sites to create the occupancy
        string and also determine the orbit mappings to evaluate the
        corresponding cluster functions.

        Args:
            structure (Structure):
                Structure to compute correlation from
            normalized (bool):
                return the correlation vector normalized by the prim cell size.
                In theory correlation vectors are always normalized, but
                getting them without normalization allows to compute the
                "extensive" values.
            scmatrix (ndarray): optional
                supercell matrix relating the prim structure to the given
                structure. Passing this if it has already been matched will
                make things much quicker. You are responsible that the
                supercell matrix is correct.
            site_mapping (list): optional
                Site mapping as obtained by
                :code:`StructureMatcher.get_mapping`
                such that the elements of site_mapping represent the indices
                of the matching sites to the prim structure. If you pass this
                option, you are fully responsible that the mappings are correct!

        Returns:
            array: correlation vector for given structure
        """
        if scmatrix is None:
            scmatrix = self.scmatrix_from_structure(structure)

        occu = self.occupancy_from_structure(
            structure, scmatrix=scmatrix, site_mapping=site_mapping, encode=True
        )
        occu = np.array(occu, dtype=int)

        # Create a list of tuples with necessary information to compute corr
        mappings = self.supercell_orbit_mappings(scmatrix)
        orbit_list = [
            (
                orbit.bit_id,
                orbit.flat_tensor_indices,
                orbit.flat_correlation_tensors,
                cluster_indices,
            )
            for cluster_indices, orbit in zip(mappings, self.orbits)
        ]
        corr = corr_from_occupancy(occu, self.num_corr_functions, orbit_list)
        size = self.num_prims_from_matrix(scmatrix)

        if self.external_terms:
            supercell = self.structure.copy()
            supercell.make_supercell(scmatrix)
            extras = [
                term.value_from_occupancy(occu, supercell) / size
                for term in self._external_terms
            ]
            corr = np.concatenate([corr, *extras])

        if not normalized:
            corr *= size

        return corr

    def refine_structure(self, structure, scmatrix=None, site_mapping=None):
        """Refine a (relaxed) structure.

        Refine a (relaxed) structure to a perfect supercell structure of the
        the prim structure (aka the corresponding "unrelaxed" structure).

        Args:
            structure (Structure):
                structure to refine to a perfect multiple of the prim
            scmatrix (ndarray): optional
                supercell matrix relating the prim structure to the given
                structure. Passing this if it has already been matched will
                make things much quicker. You are responsible for correctness.
            site_mapping (list): optional
                site mapping as obtained by StructureMatcher.get_mapping
                such that the elements of site_mapping represent the indices
                of the matching sites to the prim structure. Again you are
                responsible.

        Returns:
             Structure: the refined structure
        """
        if scmatrix is None:
            scmatrix = self.scmatrix_from_structure(structure)

        occu = self.occupancy_from_structure(
            structure, scmatrix=scmatrix, site_mapping=site_mapping
        )

        supercell_structure = self.structure.copy()
        supercell_structure.make_supercell(scmatrix)

        sites = []
        for specie, site in zip(occu, supercell_structure):
            if not isinstance(specie, Vacancy):  # skip vacancies
                site = PeriodicSite(
                    specie, site.frac_coords, supercell_structure.lattice
                )
                sites.append(site)
        return Structure.from_sites(sites)

    def occupancy_from_structure(
        self, structure, scmatrix=None, site_mapping=None, encode=False
    ):
        """Occupancy string for a given structure.

        Returns a list of occupancies of each site in the structure in the
        appropriate order set implicitly by the supercell matrix that is found.

        This function is used as input to compute correlation vectors for the
        given structure.

        This function is also useful to obtain an initial occupancy for a Monte
        Carlo simulation. (Make sure that the same supercell matrix is being
        used here as in the instance of the processor class for the simulation.
        Although it is recommended to use the similar function in Processor
        classes.)

        Args:
            structure (Structure):
                structure to obtain a occupancy string for
            scmatrix (array): optional
                supercell matrix relating the given structure and the
                primitive structure. If you pass the supercell, you fully are
                responsible that it is the correct one! This prevents running
                the _scmatcher (supercell structure matcher)
            site_mapping (list): optional
                site mapping as obtained by StructureMatcher.get_mapping
                such that the elements of site_mapping represent the indices
                of the matching sites to the prim structure. If you pass this
                option, you are fully responsible that the mappings are correct!
                This prevents running _site_matcher to get the mappings.
            encode (bool): optional
                if True, the occupancy string will have the index of the species
                in the expansion structure site spaces, rather than the
                species itself.

        Returns:
            list: occupancy string for structure.
        """
        if scmatrix is None:
            scmatrix = self.scmatrix_from_structure(structure)

        supercell = self.structure.copy()
        supercell.make_supercell(scmatrix)

        if site_mapping is None:
            site_mapping = self.structure_site_mapping(supercell, structure)

        occu = []  # np.zeros(len(self.supercell_structure), dtype=np.int)

        for i, allowed_species in enumerate(get_allowed_species(supercell)):
            # rather than starting with all vacancies and looping
            # only over mapping, explicitly loop over everything to
            # catch vacancies on improper sites
            if i in site_mapping:
                spec = structure[site_mapping.index(i)].specie
            else:
                spec = Vacancy()
            if spec not in allowed_species:
                raise StructureMatchError(
                    "A site in given structure has an  unrecognized species " f"{spec}."
                )
            if encode:
                occu.append(allowed_species.index(spec))
            else:
                occu.append(spec)
        return occu

    def scmatrix_from_structure(self, structure):
        """Get supercell matrix from a given structure.

        Obtain the supercell structure matrix to convert the prim structure
        to a supercell equivalent to given structure.

        Args:
            structure (Structure):
                a pymatgen Structure.

        Returns:
            ndarray: matrix relating given structure and prim structure.
        """
        scmatrix = self._sc_matcher.get_supercell_matrix(structure, self.structure)
        if scmatrix is None:
            raise StructureMatchError("Supercell could not be found from structure")
        if np.linalg.det(scmatrix) < 0:
            scmatrix *= -1
        return scmatrix

    def supercell_orbit_mappings(self, scmatrix):
        """Get orbit mappings for a structure from supercell of prim.

        Return the orbit mappings for a specific supercell of the prim
        structure represented by the given matrix

        Args:
            scmatrix (array):
                array relating a supercell with the primitive matrix

        Returns:
            list of tuples:
                (orbit, indices) list of tuples with orbits and the
                site indices for all equivalent orbits in a supercell obtained
                from the given matrix.
        """
        # np.arrays are not hashable and can't be used as dict keys.
        scmatrix = np.array(scmatrix)
        # so change them into a tuple of sorted tuples for unique keys.
        scm = tuple(sorted(tuple(s.tolist()) for s in scmatrix))
        indices = self._supercell_orb_inds.get(scm)

        if indices is None:
            indices = self._gen_orbit_indices(scmatrix)
            self._supercell_orb_inds[scm] = indices

        return indices

    def get_aliased_orbits(self, sc_matrix):
        """Get the aliased orbits for a given supercell shape.

        Detect the orbits that will be aliased due to translational symmetry imposed by
        the supercell lattice. Orbits i and j are aliased when a geometric cluster in
        orbit i is identically mapped to another geometric cluster in orbit j.
        It can be shown through a group theoretical argument that any cluster in orbit i
        then must be identical to a corresponding cluster in orbit j.

        The implication of aliasing is that correlation functions of these orbits will
        evaluate to the same value, leading to feature matrix rank deficiency and
        potentially unphysical ECI.

        This method will detect most cases of orbit degeneracy, but not some edge cases.

        Args:
            sc_matrix: (array):
                array relating a supercell with the primitive matrix

        Returns:
            list of tuples:
                (orbit_id i, orbit_id j, ...) list of tuples containing the orbits
                that are aliased.

        """
        sc_orb_map = self.supercell_orbit_mappings(sc_matrix)
        aliased_orbits = []
        for orb_i, orb_map_i in enumerate(sc_orb_map):
            orb_i_id = orb_i + 1
            aliased = False
            orbit_i_aliased = [orb_i_id]
            sorted_orb_map_i = {tuple(sorted(c_map)) for c_map in orb_map_i}

            for orb_j, orb_map_j in enumerate(sc_orb_map):
                if orb_i == orb_j:
                    continue
                orb_j_id = orb_j + 1
                sorted_orb_map_j = {tuple(sorted(c_map)) for c_map in orb_map_j}

                if sorted_orb_map_i == sorted_orb_map_j:
                    aliased = True
                    orbit_i_aliased.append(orb_j_id)

            orbit_i_aliased = tuple(sorted(orbit_i_aliased))
            if aliased:
                aliased_orbits.append(orbit_i_aliased)

        aliased_orbits = sorted(list(set(aliased_orbits)), key=lambda x: x[0])
        return aliased_orbits

    def change_site_bases(self, new_basis, orthonormal=False):
        """Change the type of site basis used in the site basis functions.

        Args:
            new_basis (str):
                name of new basis for all site bases
            orthonormal (bool):
                option to orthonormalize all new site basis sets
        """
        for orbit in self.orbits:
            orbit.transform_site_bases(new_basis, orthonormal)

    def remove_orbits(self, orbit_ids):
        """Remove whole orbits by their ids.

        Removes orbits from cluster spaces. It is helpful to print a
        ClusterSubspace or ClusterExpansion to obtain orbit ids. After removing
        orbits, orbit ID's and orbit bit ID's are re-assigned.

        This is useful to prune a ClusterExpansion by removing orbits with
        small associated coefficients or ECI. Note that this will remove a full
        orbit, which for the case of sites with only two species is the same as
        removing a single correlation vector element (only one ECI). For cases
        with sites having more than 2 species allowed per site there is more
        than one orbit functions (for all the possible bit orderings or function-
        labeled orbit configurations) and removing an orbit will remove more than
        one element in the correlation vector.

        Args:
            orbit_ids (list):
                list of orbit ids to be removed
        """
        if min(orbit_ids) < 0:
            raise ValueError("Index out of range. Negative inds are not allowed.")

        if min(orbit_ids) == 0:
            raise ValueError(
                "The empty orbit can not be removed. \n If you really want to "
                "do this remove the first column in your feature matrix before"
                " fitting."
            )
        if max(orbit_ids) > self.num_orbits - 1:
            raise ValueError(
                "Index out of range. " "Total number of orbits is: {self.num_orbits}"
            )

        for size, orbits in self._orbits.items():
            self._orbits[size] = [
                orbit for orbit in orbits if orbit.id not in orbit_ids
            ]

        self._assign_orbit_ids()  # Re-assign ids
        # Clear the cached supercell orbit mappings
        self._supercell_orb_inds = {}

    def remove_corr_functions(self, corr_ids):
        """Remove correlation functions by their ID's.

        This allows more granular removal of terms involved in fitting/evaluating a
        cluster expansion. Similar to remove_orbits this is useful to prune a cluster
        expansion and actually allows to remove a single term (ie one with small
        associated coefficient/ECI).

        This procedure is perfectly well posed mathematically. The resultant
        CE is still a valid function of configurations with all the necessary
        symmetries from the underlying structure. It is also practically justified
        if we allow "in group" orbit eci sparsity...which everyone in the field
        does anyway. In terms of physical/chemical interpretation it is not
        obvious what it means to remove certain combinations of an n-body
        interaction term, and not the whole term itself...so tread lighlty with your
        model interpretations.

        Args:
            corr_ids (list):
                list of correlation function ids to remove
        """
        empty_orbit_ids = []
        corr_ids = np.array(corr_ids, dtype=int)

        for orbit in self.orbits:
            first_id = orbit.bit_id
            last_id = orbit.bit_id + len(orbit)
            to_remove = corr_ids[corr_ids >= first_id]
            to_remove = to_remove[to_remove < last_id] - first_id
            if to_remove.size > 0:
                try:
                    orbit.remove_bit_combos_by_inds(to_remove)
                except RuntimeError:
                    empty_orbit_ids.append(orbit.id)
                    warnings.warn(
                        "All bit combos have been removed from orbit with id "
                        f"{orbit.id}. This orbit will be fully removed."
                    )

        if empty_orbit_ids:
            self.remove_orbits(empty_orbit_ids)
        else:
            self._assign_orbit_ids()  # Re-assign ids

    def copy(self):
        """Deep copy of instance."""
        return ClusterSubspace.from_dict(self.as_dict())

    def structure_site_mapping(self, supercell, structure):
        """Get structure site mapping.

        Returns the mapping between sites in the given structure and a prim
        supercell of the corresponding size.

        Args:
            supercell (Structure):
                supercell of prim structure with same size as other structure.
            structure (Structure):
                Structure to obtain site mappings to supercell of prim
        Returns:
            list: site mappings of structure to supercell
        """
        mapping = self._site_matcher.get_mapping(supercell, structure)
        if mapping is None:
            raise StructureMatchError("Mapping could not be found from structure.")
        return mapping.tolist()

    def get_sub_orbits(self, orbit_id, level=1, min_size=1):
        """Get sub orbits of the orbit for the corresponding orbit_id.

        Args:
            orbit_id (int):
                id of orbit to get sub orbit id for
            level (int): optional
                how many levels down to look for suborbits. If all suborbits
                are needed make level large enough or set to None.
            min_size (int): optional
                minimum size of clusters in sub orbits to include

        Returns:
            list of ints: list containing ids of suborbits
        """
        if orbit_id == 0:
            return []
        size = len(self.orbits[orbit_id - 1].base_cluster)
        if level is None or level < 0 or size - level - 1 < 0:
            stop = 0
        elif min_size > size - level:
            stop = min_size - 1
        else:
            stop = size - level - 1

        search_sizes = range(size - 1, stop, -1)
        return [
            orbit
            for s in search_sizes
            for orbit in self._orbits[s]
            if self.orbits[orbit_id - 1].is_sub_orbit(orbit)
        ]

    def get_sub_function_ids(self, corr_id, level=1, min_size=1):
        """Get the bit combo ids of all sub correlation functions.

        A sub correlation function of a given correlation function means that
        the sub correlation function is a factor of the correlation function
        (with the additional requirement of acting over the sites in sub
        clusters of the clusters over which the given corr function acts on).

        In other words, think of it as an orbit of function-labeled subclusters
        of a given orbit of function-labeled clusters...a mouthful...

        Args:
            corr_id (int):
                id of orbit to get sub orbit id for
            level (int): optional
                how many levels down to look for suborbits. If all suborbits
                are needed make level large enough or set to None.
            min_size (int): optional
                minimum size of clusters in sub orbits to include

        Returns:
            list of ints: list containing ids of sub correlation functions
        """
        if corr_id == 0:
            return []

        orbit = self.orbits[self.function_orbit_ids[corr_id] - 1]
        bit_combo = orbit.bit_combos[corr_id - orbit.bit_id]

        sub_fun_ids = []
        for sub_orbit in self.get_sub_orbits(orbit.id, level=level, min_size=min_size):
            inds = orbit.sub_orbit_mappings(sub_orbit)
            for i, sub_bit_combo in enumerate(sub_orbit.bit_combos):
                if np.any(np.all(sub_bit_combo[0] == bit_combo[:, inds], axis=2)):
                    sub_fun_ids.append(sub_orbit.bit_id + i)

        return sub_fun_ids

    def _assign_orbit_ids(self):
        """Assign unique id's to orbit.

        Assign unique id's to each orbit based on all its orbit functions and
        all clusters in the prim structure that are in each orbit.
        """
        counts = (1, 1, 1)
        for key in sorted(self._orbits.keys()):
            for orbit in self._orbits[key]:
                counts = orbit.assign_ids(*counts)

        self.num_orbits = counts[0]
        self.num_corr_functions = counts[1]
        self.num_clusters = counts[2]

    @staticmethod
    def _gen_orbits_from_cutoffs(
        exp_struct, cutoffs, symops, basis, orthonorm, use_conc
    ):
        """Generate orbits from diameter cutoffs.

        The diameter of a cluster is the maximum distance between any two
        sites in the cluster.

        Generates dictionary of {size: [Orbits]} given a dictionary of maximal
        cluster diameters and symmetry operations to apply (not necessarily all
        the symmetries of the expansion_structure).

        Args:
            exp_struct (Structure):
                Structure with all sites that have partial occupancy.
            cutoffs (dict):
                dict of cutoffs for cluster diameters {size: cutoff}.
                Cutoff diameters must decrease with cluster size, otherwise
                algorithm can not guarantee completeness of cluster subspace.
                Adding a cutoff term for point terms, ie {1: 0} is useful
                to exclude point terms any other value for the cutoff will
                simply be ignored.
            symops (list of SymmOps):
                list of symmetry operations for structure
            basis (str):
                name identifying site basis set to use.
            orthonorm (bool):
                wether to ensure orthonormal basis set.
            use_conc (bool):
                If true the concentrations in the prim structure sites will be
                used as the measure to orthormalize site bases.
        Returns:
            dict: {size: list of Orbits within diameter cutoff}
        """
        try:
            if cutoffs.pop(1) is None:
                if len(cutoffs) != 0:
                    raise ValueError(
                        f"Unable to generate clusters of higher order "
                        f" {cutoffs} if point terms are excluded."
                    )
                return {}
        except KeyError:
            pass

        site_spaces = get_site_spaces(exp_struct, include_measure=use_conc)
        site_bases = tuple(
            basis_factory(basis, site_space) for site_space in site_spaces
        )
        if orthonorm:
            for s_basis in site_bases:
                s_basis.orthonormalize()

        orbits = {}
        nbits = np.array([len(b) - 1 for b in site_spaces])

        # Generate singlet/point orbits
        orbits[1] = ClusterSubspace._gen_point_orbits(
            exp_struct, site_bases, nbits, symops
        )

        if len(cutoffs) == 0:  # return singlets only if no cutoffs provided
            return orbits

        orbits.update(
            ClusterSubspace._gen_multi_orbits(
                orbits[1], exp_struct, cutoffs, site_bases, nbits, symops
            )
        )
        return orbits

    @staticmethod
    def _gen_point_orbits(exp_struct, site_bases, nbits, symops):
        """Generate point orbits.

        Args:
            nbits (ndarray):
                array with total values for function indices per site.
            exp_struct (Structure):
                expansion structure, disordered sites only.
            site_bases (list of DiscreteBasis):
                list of site basis for each site in the expansion structure.
            symops (list of SymmOp):
                lists of symmetry operations of the underlying structure.

        Returns:
            list of Orbits:
                list of point orbits.
        """
        pt_orbits = []
        for nbit, site, sbasis in zip(nbits, exp_struct, site_bases):
            # Coordinates of point terms must stay in [0, 1] to guarantee
            # correct math of the following algorithm.
            new_orbit = Orbit(
                [np.mod(site.frac_coords, 1)],
                exp_struct.lattice,
                [list(range(nbit))],
                [sbasis],
                symops,
            )
            if new_orbit not in pt_orbits:
                pt_orbits.append(new_orbit)

        pt_orbits = sorted(
            pt_orbits,
            key=lambda x: (np.round(x.base_cluster.diameter, 6), -x.multiplicity),
        )
        return pt_orbits

    @staticmethod
    def _gen_multi_orbits(point_orbits, exp_struct, cutoffs, site_bases, nbits, symops):
        """Generate point orbits.

        Args:
            point_orbits (list of Orbit):
                list of point orbits.
            exp_struct (Structure):
                expansion structure, disordered sites only.
            cutoffs (dict):
                dict of cutoffs for cluster diameters {size: cutoff}.
                Cutoff diameters must decrease with cluster size, otherwise
                algorithm can not guarantee completeness of cluster subspace.
                Adding a cutoff term for point terms, ie {1: None} is useful
                to exclude point terms any other value for the cutoff will
                simply be ignored.
            site_bases (list of DiscreteBasis):
                list of site basis for each site in the expansion structure.
            nbits (ndarray):
                array with total values for function indices per site.
            symops (list of SymmOp):
                lists of symmetry operations of the underlying structure.

        Returns:
            dict:
                {size: list of Orbits within diameter cutoff}
        """
        # Vector sum of a, b, c divided by 2.
        # diameter + max_lp gives maximum possible distance from
        # [0.5, 0.5, 0.5] prim centoid to a point in all enumerable
        # clusters. Add SITE_TOL as a numerical tolerance grace.
        orbits = {1: point_orbits}
        max_lp = np.linalg.norm(exp_struct.lattice.matrix.sum(axis=0)) / 2
        max_lp += SITE_TOL
        for size, diameter in sorted(cutoffs.items()):
            new_orbits = []
            neighbors = exp_struct.get_sites_in_sphere(
                [0.5, 0.5, 0.5], diameter + max_lp, include_index=True
            )
            for orbit in orbits[size - 1]:
                if orbit.base_cluster.diameter > diameter:
                    continue

                for neighbor in neighbors:
                    if is_coord_subset(
                        [neighbor.frac_coords],
                        orbit.base_cluster.frac_coords,
                        atol=SITE_TOL,
                    ):
                        continue

                    new_sites = np.concatenate(
                        [orbit.base_cluster.frac_coords, [neighbor.frac_coords]]
                    )
                    new_orbit = Orbit(
                        new_sites,
                        exp_struct.lattice,
                        orbit.bits + [list(range(nbits[neighbor.index]))],
                        orbit.site_bases + [site_bases[neighbor.index]],
                        symops,
                    )

                    if new_orbit.base_cluster.diameter > diameter + 1e-8:
                        continue

                    if new_orbit not in new_orbits:
                        new_orbits.append(new_orbit)

            if len(new_orbits) > 0:
                orbits[size] = sorted(
                    new_orbits,
                    key=lambda x: (
                        np.round(x.base_cluster.diameter, 6),
                        -x.multiplicity,
                    ),
                )
        return orbits

    def _gen_orbit_indices(self, scmatrix):
        """Find all the site indices associated with each orbit in structure.

        The structure corresponding to the given supercell matrix w.r.t prim.
        """
        supercell = self.structure.copy()
        supercell.make_supercell(scmatrix)
        prim_to_supercell = np.linalg.inv(scmatrix)
        supercell_fcoords = np.array(supercell.frac_coords)

        pts = lattice_points_in_supercell(scmatrix)
        orbit_indices = []
        for orbit in self.orbits:
            prim_fcoords = np.array([c.frac_coords for c in orbit.clusters])
            fcoords = np.dot(prim_fcoords, prim_to_supercell)
            # tcoords contains all the coordinates of the symmetrically
            # equivalent clusters the indices are: [equivalent cluster
            # (primitive cell), translational image, index of site in cluster,
            # coordinate index]
            tcoords = fcoords[:, None, :, :] + pts[None, :, None, :]
            tcs = tcoords.shape
            inds = coord_list_mapping_pbc(
                tcoords.reshape((-1, 3)), supercell_fcoords, atol=SITE_TOL
            ).reshape((tcs[0] * tcs[1], tcs[2]))
            # orbit_ids holds orbit, and 2d array of index groups that
            # correspond to the orbit
            # the 2d array may have some duplicates. This is due to
            # symetrically equivalent groups being matched to the same sites
            # (eg in simply cubic all 6 nn interactions will all be [0, 0]
            # indices. This multiplicity disappears as supercell_structure size
            # increases, so I haven't implemented a more efficient method
            orbit_indices.append(inds)

        return orbit_indices

    def __eq__(self, other):
        """Check equality between cluster subspaces."""
        if not isinstance(other, ClusterSubspace):
            return False
        if other.num_corr_functions != self.num_corr_functions:
            return False
        if len(self.external_terms) != len(other.external_terms):
            return False
        if not all(
            isinstance(t1, type(t2))
            for t1, t2 in zip(other.external_terms, self.external_terms)
        ):
            return False
        # does not check if basis functions are the same.
        return all(o1 == o2 for o1, o2 in zip(other.orbits, self.orbits))

    def __contains__(self, orbit):
        """Check if subspace contains orbit."""
        return orbit in self.orbits

    def __len__(self):
        """Get number of correlation functions and ext terms in subspace."""
        return self.num_corr_functions + len(self.external_terms)

    def __str__(self):
        """Convert class into pretty string for printing."""
        outs = [
            f"Basis/Orthogonal/Orthonormal : {self.basis_type}/{self.basis_orthogonal}/"
            "{self.basis_orthonormal}",
            f"       Unit Cell Composition : {self.structure.composition}",
            f"            Number of Orbits : {self.num_orbits}",
            f"No. of Correlation Functions : {self.num_corr_functions}",
            "             Cluster Cutoffs : "
            f"{', '.join('{}: {:.2f}'.format(s, c) for s, c in self.cutoffs.items())}",
            f"              External Terms : {self.external_terms}",
            "Orbit Summary",
            " ------------------------------------------------------------------------",
            " |  ID     Degree    Cluster Diameter    Multiplicity    No. Functions  |",
            " |   0       0             NA                 0                1        |",
        ]
        for degree, orbits in self.orbits_by_size.items():
            for orbit in orbits:
                outs.append(
                    f" |{orbit.id:^7}{degree:^10}{orbit.base_cluster.diameter:^20.4f}"
                    f"{orbit.multiplicity:^16}{len(orbit):^17}|"
                )
        outs.append(
            " ------------------------------------------------------------------------"
        )
        return "\n".join(outs)

    def __repr__(self):
        """Return a summary of subspace."""
        outs = [
            "Cluster Subspace Summary",
            f"Basis/Orthogonal/Orthonormal : {self.basis_type}/{self.basis_orthogonal}/"
            f"{self.basis_orthonormal}",
            f"Unit Cell Composition : {self.structure.composition}",
            f"Number of Orbits : {self.num_orbits}   "
            f"No. of Correlation Functions : {self.num_corr_functions}",
            "Cluster Cutoffs : "
            f"{', '.join('{}: {:.2f}'.format(s, c) for s, c in self.cutoffs.items())}",
            f"External Terms : {self.external_terms}",
        ]
        return "\n".join(outs)

    @classmethod
    def from_dict(cls, d):
        """Create ClusterSubspace from an MSONable dict."""
        symops = [SymmOp.from_dict(so_d) for so_d in d["symops"]]
        orbits = {
            int(s): [Orbit.from_dict(o) for o in v] for s, v in d["orbits"].items()
        }
        structure = Structure.from_dict(d["structure"])
        exp_structure = Structure.from_dict(d["expansion_structure"])
        sc_matcher = StructureMatcher.from_dict(d["sc_matcher"])
        site_matcher = StructureMatcher.from_dict(d["site_matcher"])
        cluster_subspace = cls(
            structure=structure,
            expansion_structure=exp_structure,
            orbits=orbits,
            symops=symops,
            supercell_matcher=sc_matcher,
            site_matcher=site_matcher,
        )

        # attempt to recreate external terms. This can be much improved if
        # a base class is used.
        # TODO update this using instances of BasePairTerm when the time comes
        for term in d["external_terms"]:
            try:
                module = import_module(term["@module"])
                term_class = getattr(module, term["@class"])
                cluster_subspace.add_external_term(term_class.from_dict(term))
            except AttributeError:
                warnings.warn(
                    f"{term['@class']} was not found in {term['@module']}. You"
                    f" will need to add this yourself.",
                    RuntimeWarning,
                )
            except ImportError:
                warnings.warn(
                    f"Module {term['@module']} for class {term['@class']} was "
                    f"not found. You will have to add this yourself.",
                    ImportWarning,
                )
        # re-create supercell orb inds cache
        _supercell_orb_inds = {}
        for scm, indices in d["_supercell_orb_inds"]:
            scm = tuple(tuple(s) for s in scm)
            if isinstance(indices[0][0], int) and isinstance(indices[0][1], list):
                warnings.warn(
                    "This ClusterSubspace was created with a previous version "
                    "of smol. Please resave it to avoid this warning.",
                    FutureWarning,
                )
                _supercell_orb_inds[scm] = [np.array(ind) for o_id, ind in indices]
            else:
                _supercell_orb_inds[scm] = [np.array(ind) for ind in indices]
        cluster_subspace._supercell_orb_inds = _supercell_orb_inds
        return cluster_subspace

    def as_dict(self):
        """
        Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        # modify cached sc orb inds so it can be serialized
        _supercell_orb_inds = [
            (scm, [ind.tolist() for ind in orb_inds])
            for scm, orb_inds in self._supercell_orb_inds.items()
        ]

        cs_dict = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "structure": self.structure.as_dict(),
            "expansion_structure": self.expansion_structure.as_dict(),
            "symops": [so.as_dict() for so in self.symops],
            "orbits": {s: [o.as_dict() for o in v] for s, v in self._orbits.items()},
            "sc_matcher": self._sc_matcher.as_dict(),
            "site_matcher": self._site_matcher.as_dict(),
            "external_terms": [et.as_dict() for et in self.external_terms],
            "_supercell_orb_inds": _supercell_orb_inds,
        }
        return cs_dict


def invert_mapping(mapping):
    """Invert a mapping table from forward to backward, vice versa.

    Args:
        mapping (list of lists):
            List of sublists, each contains integer indices, indicating
            a foward mapping from the current sublist index to the indices
            in the sublist.

    Returns:
        lists of lists: Inverted mapping table containing backward mapping.
        Same format as input.
    """
    inv_mapping = [[] for _ in range(len(mapping))]

    for i in range(len(mapping) - 1, -1, -1):
        values_list = mapping[i]
        for value in values_list:
            if i not in inv_mapping[value]:
                inv_mapping[value].append(i)

    return inv_mapping


def get_complete_mapping(mapping):
    """Get a complete mapping from a 1 level mapping.

    If we allow transferability between mapping linkages, there would be
    a full mapping containing all linkages at all connectivity level.
    Using this function, you can get a full mapping from an incomplete
    mapping.

    NOTE: Since complete hierarchy is not very useful for actual CE fit, we
    will not include it as an attribute of ClusterSubspace.

    Args:
        mapping (list of lists):
             List of sublists, each contains integer indices, indicating
            a foward mapping from the current sublist index to the indices
            in the sublist.

    Returns:
        list of lists: Full mapping table containing forward mapping, but with
        all connectivity levels. Same format as input
    """
    all_level_mapping = deepcopy(mapping)

    for i in range(len(mapping) - 1, -1, -1):
        next_values_list = mapping[i]

        while len(next_values_list) > 0:
            for next_value in next_values_list:
                if next_value not in all_level_mapping[i]:
                    all_level_mapping[i].append(next_value)
            next_values_list_new = []
            for next_value in next_values_list:
                for nn_value in mapping[next_value]:
                    if nn_value not in next_values_list_new:
                        next_values_list_new.append(nn_value)
            next_values_list = next_values_list_new

    return all_level_mapping


class PottsSubspace(ClusterSubspace):
    """PottsSubspace represents a subspace of functions using only indicator functions.

    A PottsSubspace is a related model to a standard ClusterSubspace used to build
    a standard cluster expansion. The only difference is that the single site functions
    for any orbit are only made up of indicator functions (and there is no constant
    function). As such it is more closely related to a generalized Ising model or better
    yet an extension of the q-state Potts Model (hence the name).

    The orbit functions in a PottsSubspace represent the concentrations of all possible
    decorations (species occupancies) of clusters in the given orbit. Similar to a
    ClusterSubspace with site indicator basis functions. But in contrast, the Potts
    subspace includes the concentration of all possible decorations (minus 1).

    Although quite similar to a ClusterSubspace, there is no mathematical
    formalism guaranteeing that the orbit basis functions generated in a
    PottsSubspace are a linear independent set spanning configuration space.
    Actually if all orbits up to any size (or infinite size) are included,
    the corresponding orbit function set is an overcomplete/ highly redundant.

    A PottsSubspace can be created directly with a ClusterSubspace object
    by using the constructor and providing the appropriately constructed orbits,
    but that is a bit more painful that just using the class method here.
    """

    def __init__(
        self,
        structure,
        expansion_structure,
        symops,
        orbits,
        without_last_cluster=True,
        supercell_matcher=None,
        site_matcher=None,
        **matcher_kwargs,
    ):
        """Initialize a PottsSubspace.

        You rarely will need to create a ClusterSubspace using the main
        constructor.
        Look at the class method :code:`from_cutoffs` for the "better" way to
        instantiate a ClusterSubspace.

        Args:
            structure (Structure):
                Structure to define the cluster space. Typically the primitive
                cell. Includes all species regardless of partial occupation.
            expansion_structure (Structure):
                Structure including only sites that will be included in the
                Cluster space. (only those with partial occupancy)
            symops (list of Symmop):
                list of Symmops for the given structure.
            orbits (dict): {size: list of Orbits}
                Dictionary with size (number of sites) as keys and list of
                Orbits as values.
            without_last_cluster (bool): optional
                whether last cluster labeling is removed from each orbit.
            supercell_matcher (StructureMatcher): (optional)
                StructureMatcher used to find supercell matrices
                relating the prim structure to other structures. If you pass
                this directly you should know how to set the matcher up,
                otherwise matching your relaxed structures can fail, alot.
            site_matcher (StructureMatcher): (optional)
                StructureMatcher used to find site mappings
                relating the sites of a given structure to an appropriate
                supercell of the prim structure . If you pass this directly you
                should know how to set the matcher up, otherwise matching your
                relaxed structures can fail, alot.
            matcher_kwargs:
                ltol, stol, angle_tol, supercell_size: parameters to pass
                through to the StructureMatchers. Structures that don't match
                to the primitive cell under these tolerances won't be included
                in the expansion. Easiest option for supercell_size is usually
                to use a species that has a constant amount per formula unit.
                See pymatgen documentation of :class:`StructureMatcher` for
                more details.
        """
        self._wo_last_cluster = without_last_cluster
        super().__init__(
            structure,
            expansion_structure,
            symops,
            orbits,
            supercell_matcher,
            site_matcher,
            **matcher_kwargs,
        )

    @classmethod
    def from_cutoffs(
        cls,
        structure,
        cutoffs,
        remove_last_cluster=False,
        supercell_matcher=None,
        site_matcher=None,
        **matcher_kwargs,
    ):
        """Create a PottsSubspace from diameter cutoffs.

        Creates a :class:`PottsSubspace` with orbits of the given size and
        diameter smaller than or equal to the given value. The diameter of an
        orbit is the maximum distance between any two sites of a cluster of
        that orbit.

        Args:
           structure (Structure):
               disordered structure to build a cluster expansion for.
               Typically the primitive cell
           cutoffs (dict):
               dict of {cluster_size: diameter cutoff}. Cutoffs should be
               strictly decreasing. Typically something like {2:5, 3:4}.
               The empty orbit is always included. Singlets are by default
               included, with the exception below.
               To obtain a subspace with only an empty and singlet terms use
               an empty dict {}, or {1: 1}. Adding a cutoff term for point
               terms, i.e. {1: None} is useful to exclude point terms, any other
               value for the cutoff will simply be ignored.
           remove_last_cluster (bool): optional
               if True, will remove the last cluster labeling (decoration)
               from each orbit. Since sum of corr for all labelings = 1,
               removing the last is similar to working in concentration space.
           supercell_matcher (StructureMatcher): (optional)
               StructureMatcher used to find supercell matrices
               relating the prim structure to other structures. If you pass
               this directly you should know how to set the matcher up,
               otherwise matching your relaxed structures will fail, alot.
           site_matcher (StructureMatcher): (optional)
               StructureMatcher used to find site mappings
               relating the sites of a given structure to an appropriate
               supercell of the prim structure . If you pass this directly you
               should know how to set the matcher up, otherwise matching your
               relaxed structures will fail, alot.
           matcher_kwargs:
               ltol, stol, angle_tol, supercell_size: parameters to pass
               through to the StructureMatchers. Structures that don't match
               to the primitive cell under these tolerances won't be included
               in the expansion. Easiest option for supercell_size is usually
               to use a species that has a constant amount per formula unit.

        Returns:
           PottsSubSpace
        """
        # get symmetry operations of prim structure.
        symops = SpacegroupAnalyzer(structure).get_symmetry_operations()
        # get the active sites (partial occupancy) to expand over.
        sites_to_expand = [
            site
            for site in structure
            if site.species.num_atoms < 0.99 or len(site.species) > 1
        ]
        expansion_structure = Structure.from_sites(sites_to_expand)
        # get orbits within given cutoffs
        orbits = cls._gen_orbits_from_cutoffs(
            expansion_structure, cutoffs, symops, remove_last_cluster
        )
        return cls(
            structure=structure,
            expansion_structure=expansion_structure,
            symops=symops,
            orbits=orbits,
            without_last_cluster=remove_last_cluster,
            supercell_matcher=supercell_matcher,
            site_matcher=site_matcher,
            **matcher_kwargs,
        )

    def get_function_decoration(self, index):
        """Get the decoration/labeling of a specific orbit function.

        When using an indicator site basis there is 1 to 1 equivalence between
        correlation functions and species decorations.

        Args:
            index (int):
                index of orbit function in correlation vector

        Returns:
            list of tuples: list of tuples of symmetrically equivalent
            Species/Elements.
        """
        orbit = self.orbits[self.function_orbit_ids[index] - 1]
        decorations = [
            tuple(list(orbit.site_spaces[i])[b] for i, b in enumerate(bits))
            for bits in orbit.bit_combos[index - orbit.bit_id]
        ]
        return decorations

    def get_orbit_decorations(self, orbit_id):
        """Get all decorations/labellings of species in an orbit.

        Args:
            orbit_id (int):
                ID of orbit

        Returns:
            list of list: list of lists of symmetrically equivalent
            Species/Elements.
        """
        bit_id = self.orbits[orbit_id - 1].bit_id
        num_combos = len(self.orbits[orbit_id - 1].bit_combos)
        return [
            self.get_function_decoration(bid)
            for bid in range(bit_id, bit_id + num_combos)
        ]

    @staticmethod
    def _gen_orbits_from_cutoffs(exp_struct, cutoffs, symops, remove_last):
        """Generate orbits from diameter cutoffs.

        Generates dictionary of orbits in the same way that the cluster
        subspace class does, except that the orbit functions (and corresponding
        bit combos) include all symmetrically distinct decorations/labelings
        of indicator functions for all allowed species (except 1 decoration
        for each orbit since this value is just 1 - sum of concentration of all
        other decorations

        Args:
            exp_struct (Structure):
                Structure with all sites that have partial occupancy.
            cutoffs (dict):
                dict of cutoffs for cluster diameters {size: cutoff}
            symops (list of SymmOps):
                list of symmetry operations for structure
            remove_last (bool):
                remove the last cluster labeling from each orbit.

        Returns:
            dict: {size: list of Orbits within diameter cutoff}
        """
        site_spaces = get_site_spaces(exp_struct)
        site_bases = tuple(IndicatorBasis(site_space) for site_space in site_spaces)
        orbits = {}
        nbits = np.array([len(b) for b in site_spaces])

        try:
            if cutoffs.pop(1) is None:
                if len(cutoffs) != 0:
                    raise ValueError(
                        f"Unable to generate clusters of higher order "
                        f" {cutoffs} if point terms are excluded."
                    )
                return {}
        except KeyError:
            pass

        # Generate singlet/point orbits
        orbits[1] = ClusterSubspace._gen_point_orbits(
            exp_struct, site_bases, nbits, symops
        )

        if len(cutoffs) == 0:  # return singlets only if no cutoffs provided
            return orbits

        orbits.update(
            ClusterSubspace._gen_multi_orbits(
                orbits[1], exp_struct, cutoffs, site_bases, nbits, symops
            )
        )

        if remove_last:
            for orbs in orbits.values():
                for orb in orbs:
                    orb.remove_bit_combos_by_inds([len(orb.bit_combos) - 1])

        return orbits

    def as_dict(self):
        """
        Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        # pylint: disable=protected-access, no-member
        pcs_dict = super().as_dict()
        pcs_dict["_wo_last_cluster"] = self._wo_last_cluster
        return pcs_dict

    @classmethod
    def from_dict(cls, d):
        """Create ClusterSubspace from an MSONable dict."""
        # pylint: disable=protected-access, no-member
        subspace = super().from_dict(d)
        subspace._wo_last_cluster = d.get("_wo_last_cluster", True)
        # remove last bit combo in all orbits
        if subspace._wo_last_cluster:
            for orbit in subspace.orbits:
                orbit.remove_bit_combos_by_inds([len(orbit.bit_combos) - 1])
        subspace._assign_orbit_ids()
        return subspace



"""Implementation of SampleContainer class.

A SampleContainer is used to store data from Monte Carlo sampling simulation.
It has some minimimal methods and properties useful to start analyzing the
samples.
"""

__author__ = "Luis Barroso-Luque"

import json
import os
import warnings
from collections import defaultdict

import numpy as np
from monty.dev import requires
from monty.json import MSONable, jsanitize

from smol.moca.sampler.kernel import Trace
from smol.moca.sublattice import Sublattice

try:
    import h5py
except ImportError:
    h5py = None


class SampleContainer(MSONable):
    """A SampleContainer class stores Monte Carlo simulation samples.

    A SampleContainer holds samples and sampling information from an MCMC
    sampling run. It is useful to obtain the raw data and minimal empirical
    properties of the underlying distribution in order to carry out further
    analysis of the MCMC sampling results.

    When getting any value from the provided attributes, the highly repeated
    args are:

        discard (int): optional
            number of samples to discard to obtain the value requested.
        thin_by (int): optional
            use every thin_by sample to obtain the value requested.
        flat (bool): optional
            if more than 1 walker is used, flattening will flatten all
            chains into one. Defaults to True.

    Attributes:
        num_sites (int):
            Size of system (usually in number of prims in supercell, but
            can be anything representative, i.e. number of sites)
        sublattices (list of Sublattice)
            Sublattices of the ensemble sampled.
        natural_parameters (ndarray):
                Array of natural parameters used in the ensemble.
        metadata (dict):
            Dictionary of metadata from the MC run that generated the samples.
    """

    def __init__(
        self,
        sublattices,
        natural_parameters,
        num_energy_coefs,
        sample_trace,
        sampling_metadata=None,
    ):
        """Initialize a sample container.

        Args:
            sublattices (list of Sublattice):
                Sublattices of the ensemble sampled.
            natural_parameters (ndarray):
                array of natural parameters used in the ensemble.
            num_energy_coefs (int):
                the number of coefficients in the natural parameters that
                correspond to the energy only.
            sample_trace (Trace):
                a trace object for the traced values during MC sampling
            sampling_metadata (Ensemble):
                sampling metadata (i.e. ensemble name, mckernel type, etc)
        """
        self.sublattices = sublattices
        self.natural_parameters = natural_parameters
        self._total_steps = 0
        self.metadata = {} if sampling_metadata is None else sampling_metadata
        self._num_energy_coefs = num_energy_coefs
        # need counter because cant use shape of arraus when allocating space
        self._nsamples = 0
        self._trace = sample_trace
        self._aux_checkpoint = None
        self._backend = None  # for streaming

    @property
    def num_samples(self):
        """Get the total number of samples."""
        return self._nsamples

    @property
    def total_mc_steps(self):
        """Return the total number of MC steps taken during sampling."""
        return self._total_steps

    @property
    def shape(self):
        """Get the shape of the samples in chain."""
        return self._trace.occupancy.shape[1:]

    @property
    def traced_values(self):
        """Get the names of traced values being sampled."""
        return self._trace.names

    def sampling_efficiency(self, discard=0, flat=True):
        """Return the sampling efficiency for chains."""
        total_accepted = self._trace.accepted[discard:].sum(axis=0)
        efficiency = total_accepted / (self._total_steps - discard)
        if flat:
            efficiency = efficiency.mean()
        return efficiency

    def get_trace_value(self, name, discard=0, thin_by=1, flat=True):
        """Get sampled values of a traced value given by name."""
        value = getattr(self._trace, name)[discard + thin_by - 1 :: thin_by]
        if flat:
            value = self._flatten(value)
        return value

    def mean_trace_value(self, name, discard=0, thin_by=1, flat=True):
        """Get mean of a traced value given by name."""
        return self.get_trace_value(name, discard, thin_by, flat).mean(axis=0)

    def trace_value_variance(self, name, discard=0, thin_by=1, flat=True):
        """Get variance of a traced value given by name."""
        return self.get_trace_value(name, discard, thin_by, flat).var(axis=0)

    def get_occupancies(self, discard=0, thin_by=1, flat=True):
        """Get the occupancy chain of the samples."""
        return self.get_trace_value("occupancy", discard, thin_by, flat)

    def get_enthalpies(self, discard=0, thin_by=1, flat=True):
        """Get the generalized enthalpy changes from samples in chain."""
        return self.get_trace_value("enthalpy", discard, thin_by, flat)

    def get_feature_vectors(self, discard=0, thin_by=1, flat=True):
        """Get the feature vector changes from samples in chain."""
        return self.get_trace_value("features", discard, thin_by, flat)

    def get_energies(self, discard=0, thin_by=1, flat=True):
        """Get the energies from samples in chain."""
        if len(self.natural_parameters) == self._num_energy_coefs:
            return self.get_enthalpies(discard, thin_by, flat)
        # otherwise we have to calculate without the additional terms...
        feature_trace = self.get_feature_vectors(discard, thin_by, flat=False)
        energies = np.expand_dims(
            np.vstack(
                [
                    np.tensordot(
                        features[:, : self._num_energy_coefs],
                        self.natural_parameters[: self._num_energy_coefs],
                        axes=([1], [0]),
                    )
                    for features in feature_trace
                ]
            ),
            axis=-1,
        )
        if flat:
            energies = self._flatten(energies)
        return energies

    def get_sublattice_compositions(self, sublattice, discard=0, thin_by=1, flat=True):
        """Get the compositions of a specific sublattice."""
        counts = self.get_sublattice_species_counts(sublattice, discard, thin_by, flat)
        return counts / len(sublattice.sites)

    def get_compositions(self, discard=0, thin_by=1, flat=True):
        """Get the compositions for each occupancy in the chain."""
        counts = self.get_species_counts(discard, thin_by, flat)
        return {spec: count / self.shape[1] for spec, count in counts.items()}

    def mean_enthalpy(self, discard=0, thin_by=1, flat=True):
        """Get the mean generalized enthalpy."""
        return self.get_enthalpies(discard, thin_by, flat).mean(axis=0)

    def enthalpy_variance(self, discard=0, thin_by=1, flat=True):
        """Get the variance in enthalpy."""
        return self.get_enthalpies(discard, thin_by, flat).var(axis=0)

    def mean_energy(self, discard=0, thin_by=1, flat=True):
        """Calculate the mean energy from samples."""
        return self.get_energies(discard, thin_by, flat).mean(axis=0)

    def energy_variance(self, discard=0, thin_by=1, flat=True):
        """Calculate the variance of sampled energies."""
        return self.get_energies(discard, thin_by, flat).var(axis=0)

    def mean_feature_vector(self, discard=0, thin_by=1, flat=True):
        """Get the mean feature vector from samples."""
        return self.get_feature_vectors(discard, thin_by, flat).mean(axis=0)

    def feature_vector_variance(self, discard=0, thin_by=1, flat=True):
        """Get the variance of feature vector elements."""
        return self.get_feature_vectors(discard, thin_by, flat).var(axis=0)

    def mean_composition(self, discard=0, thin_by=1, flat=True):
        """Get mean composition for all species regardless of sublattice."""
        comps = self.get_compositions(discard, thin_by, flat)
        return {spec: comp.mean(axis=0) for spec, comp in comps.items()}

    def composition_variance(self, discard=0, thin_by=1, flat=True):
        """Get the variance in composition of all species."""
        comps = self.get_compositions(discard, thin_by, flat)
        return {spec: comp.var(axis=0) for spec, comp in comps.items()}

    def mean_sublattice_composition(self, sublattice, discard=0, thin_by=1, flat=True):
        """Get the mean composition of a specific sublattice."""
        return self.get_sublattice_compositions(
            sublattice, discard, thin_by, flat
        ).mean(axis=0)

    def sublattice_composition_variance(
        self, sublattice, discard=0, thin_by=1, flat=True
    ):
        """Get the variance in composition of a specific sublattice."""
        return self.get_sublattice_compositions(sublattice, discard, thin_by, flat).var(
            axis=0
        )

    def get_minimum_enthalpy(self, discard=0, thin_by=1, flat=True):
        """Get the minimum energy from samples."""
        return self.get_enthalpies(discard, thin_by, flat).min(axis=0)

    def get_minimum_enthalpy_occupancy(self, discard=0, thin_by=1, flat=True):
        """Find the occupancy with minimum energy from samples."""
        inds = self.get_enthalpies(discard, thin_by, flat).argmin(axis=0)
        if flat:
            occus = self.get_occupancies(discard, thin_by, flat)[inds]
        else:
            occus = self.get_occupancies(discard, thin_by, flat)[
                inds, np.arange(self.shape[0])
            ][0]
        return occus

    def get_minimum_energy(self, discard=0, thin_by=1, flat=True):
        """Get the minimum energy from samples."""
        return self.get_energies(discard, thin_by, flat).min(axis=0)

    def get_minimum_energy_occupancy(self, discard=0, thin_by=1, flat=True):
        """Find the occupancy with minimum energy from samples."""
        inds = self.get_energies(discard, thin_by, flat).argmin(axis=0)
        if flat:
            occus = self.get_occupancies(discard, thin_by, flat)[inds]
        else:
            occus = self.get_occupancies(discard, thin_by, flat)[
                inds, np.arange(self.shape[0])
            ][0]
        return occus

    def get_species_counts(self, discard=0, thin_by=1, flat=True):
        """Get the species counts for each occupancy in the chain."""
        samples = (self.num_samples - discard) // thin_by
        shape = self.shape[0] * samples if flat else (self.shape[0], samples)
        counts = defaultdict(lambda: np.zeros(shape=shape))
        for sublattice in self.sublattices:
            subcounts = self.get_sublattice_species_counts(
                sublattice, discard, thin_by, flat
            )
            for species, count in zip(sublattice.species, subcounts.T):
                counts[species] += count
        return counts

    def get_sublattice_species_counts(
        self, sublattice, discard=0, thin_by=1, flat=True
    ):
        """Get the counts of each species in a sublattices.

        Returns:
            ndarray: where last axis is the count for each species in the
            same order as the underlying site space.
        """
        if sublattice not in self.sublattices:
            raise ValueError(
                "Sublattice provided is not recognized.\n Provide one included"
                " in the sublattices attribute of this SampleContainer."
            )
        occus = self.get_occupancies(discard, thin_by, flat=False)
        counts = np.zeros((*occus.shape[:-1], len(sublattice.site_space)))
        #  This can probably be re-written in a clean/faster way
        for i, occupancies in enumerate(occus):
            for j, occupancy in enumerate(occupancies):
                codes, count = np.unique(
                    occupancy[sublattice.sites], return_counts=True
                )
                # check for zero counts
                if len(codes) != len(sublattice.site_space):
                    missed = list(set(sublattice.encoding) - set(codes))
                    codes = np.append(codes, missed)
                    count = np.append(count, len(missed) * [0])

                original_codes = sublattice.encoding.tolist()
                order = [codes.tolist().index(code) for code in original_codes]
                counts[i][j] = count[order]  # order them accordingly
        if flat:
            counts = self._flatten(counts)
        return counts

    def save_sampled_trace(self, trace, thinned_by):
        """Save a sampled trace.

        Args:
            trace (Trace)
                Trace of sampled values
            thinned_by (int):
                the amount that the sampling was thinned by. Used to update
                the total mc iterations.
        """
        for name, value in trace.items():
            getattr(self._trace, name)[self._nsamples] = value
        self._nsamples += 1
        self._total_steps += thinned_by

    def clear(self):
        """Clear all samples from container."""
        self._total_steps = 0
        self._nsamples = 0
        for name, value in self._trace.items():
            setattr(
                self._trace, name, np.empty((0, *value.shape[1:]), dtype=value.dtype)
            )

    def allocate(self, nsamples):
        """Allocate more space in arrays for more samples."""
        for name, value in self._trace.items():
            arr = np.empty((nsamples, *value.shape[1:]), dtype=value.dtype)
            setattr(self._trace, name, np.append(value, arr, axis=0))

    def flush_to_backend(self, backend):
        """Flush current samples and trace to backend file.

        Args:
            backend (object):
                backend file object, currently only hdf5 supported.
        """
        start = backend["trace"].attrs["nsamples"]
        end = len(self._trace.occupancy) + start
        for name, value in self._trace.items():
            backend["trace"][name][start:end] = value

        backend["trace"].attrs["total_mc_steps"] += self._total_steps
        backend["trace"].attrs["nsamples"] += self._nsamples
        backend.flush()
        self._total_steps = 0
        self._nsamples = 0

    @requires(h5py is not None, "'h5py' not found. Please install it.")
    def get_backend(self, file_path, alloc_nsamples=0, swmr_mode=False):
        """Get a backend file object.

        Currently only hdf5 files supported

        Args:
            file_path (str):
                path to backend file.
            alloc_nsamples (int): optional
                number of new samples to allocate. Will only extend datasets
                if number given is larger than space left to write samples
                into.
            swmr_mode (bool): optional
                If true allows to read file from other processes. Single Writer
                Multiple Readers.

        Returns:
            h5.File object
        """
        if os.path.isfile(file_path):
            backend = self._check_backend(file_path)
            trace_grp = backend["trace"]
            available = len(trace_grp.occupancy) - trace_grp.attrs["nsamples"]
            # this probably fails since maxshape is not set.
            if available < alloc_nsamples:
                SampleContainer._grow_backend(backend, alloc_nsamples - available)
        else:
            backend = h5py.File(file_path, "w", libver="latest")
            self._init_backend(backend, alloc_nsamples)

        if swmr_mode:
            backend.swmr_mode = swmr_mode
        return backend

    def _check_backend(self, file_path):
        """Check if existing backend file is populated correctly."""
        backend = h5py.File(file_path, mode="r+", libver="latest")
        if self.shape != backend["trace"]["occupancy"].shape[1:]:
            shape = backend["trace"]["occupancy"].shape[1:]
            backend.close()
            raise RuntimeError(
                f"Backend file {file_path} has incompatible dimensions "
                f"{self.shape}, {shape}."
            )
        return backend

    def _init_backend(self, backend, nsamples):
        """Initialize a backend file."""
        sublattices = [sublatt.as_dict() for sublatt in self.sublattices]
        backend.create_dataset("sublattices", data=json.dumps(sublattices))
        backend.create_dataset("natural_parameters", data=self.natural_parameters)
        backend["natural_parameters"].attrs[
            "num_energy_coefs"
        ] = self._num_energy_coefs  # noqa
        backend.create_dataset("sampling_metadata", data=json.dumps(self.metadata))
        trace_grp = backend.create_group("trace")
        for name, value in self._trace.items():
            trace_grp.create_dataset(
                name,
                shape=(nsamples, *value.shape[1:]),
                dtype=value.dtype,
                maxshape=(None, *value.shape[1:]),
            )
        trace_grp.attrs["nsamples"] = 0
        trace_grp.attrs["total_mc_steps"] = 0

    @staticmethod
    def _grow_backend(backend, nsamples):
        """Extend space available in a backend file."""
        for name in backend["trace"]:
            backend["trace"][name].resize(nsamples, axis=0)

    @staticmethod
    def _flatten(traced_values):
        """Flatten values in trace values with multiple walkers."""
        shape_l = list(traced_values.shape[1:])
        shape_l[0] = np.prod(traced_values.shape[:2])
        return np.squeeze(traced_values.reshape(shape_l))

    def __len__(self):
        """Return the number of samples."""
        return self._nsamples

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        container_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "sublattices": [s.as_dict() for s in self.sublattices],
            "natural_parameters": jsanitize(self.natural_parameters),
            "metadata": self.metadata,
            "num_energy_coefs": self._num_energy_coefs,
            "total_mc_steps": self._total_steps,
            "nsamples": self._nsamples,
            "trace": jsanitize(self._trace.as_dict()),
            "aux_checkpoint": self._aux_checkpoint,
        }
        # TODO need to think how to generally serialize the aux checkpoint
        return container_d

    @classmethod
    def from_dict(cls, d):
        """Instantiate a SampleContainer from dict representation.

        Args:
            d (dict):
                dictionary representation.
        Returns:
            Sublattice
        """
        sublattices = [Sublattice.from_dict(s) for s in d["sublattices"]]
        trace = Trace(**{key: np.array(val) for key, val in d["trace"].items()})
        container = cls(
            sublattices,
            d["natural_parameters"],
            d["num_energy_coefs"],
            trace,
            d["metadata"],
        )
        # set the container internals
        container._total_steps = d["total_mc_steps"]
        container._nsamples = trace.occupancy.shape[0]
        container._aux_checkpoint = d["aux_checkpoint"]
        return container

    def to_hdf5(self, file_path):
        """Save SampleContainer as an HDF5 file.

        Args:
            file_path (str):
                path to file save location. If file exists and dimensions
                match samples will be appended.
        """
        # keep these to not reset writing
        total_mc_steps, nsamples = self._total_steps, self._nsamples
        backend = self.get_backend(file_path, self.num_samples)
        self.flush_to_backend(backend)
        self._total_steps, self._nsamples = total_mc_steps, nsamples
        backend.close()

    @classmethod
    @requires(h5py is not None, "'h5py' not found. Please install it.")
    def from_hdf5(cls, file_path, swmr_mode=True):
        """Instantiate a SampleContainer from an hdf5 file.

        Args:
            file_path (str):
                path to file
            swmr_mode (bool): optional
                If true allows to read file from other processes. Single Writer
                Multiple Readers.

        Returns:
            SampleContainer
        """
        with h5py.File(file_path, "r", swmr=swmr_mode) as f:
            # Check if written states matches the size of datasets
            nsamples = f["trace"].attrs["nsamples"]
            if len(f["trace"]["occupancy"]) > nsamples:
                warnings.warn(
                    f"The hdf5 file provided appears to be from an unifinished"
                    f" MC run.\n Only {nsamples} of "
                    f" {len(f['trace']['occupancy'])} samples "
                    f"have been written and will be loaded.",
                    UserWarning,
                )

            sublattices = [
                Sublattice.from_dict(s) for s in json.loads(f["sublattices"][()])
            ]
            trace = Trace(
                **{name: value[:nsamples] for name, value in f["trace"].items()}
            )
            container = cls(
                sublattices=sublattices,
                natural_parameters=f["natural_parameters"][()],
                num_energy_coefs=f["natural_parameters"].attrs[
                    "num_energy_coefs"
                ],  # noqa
                sample_trace=trace,
                sampling_metadata=json.loads(f["sampling_metadata"][()]),
            )
            container._nsamples = nsamples
            container._total_steps = f["trace"].attrs["total_mc_steps"]
        return container



"""Implementation of Sublattice class.

A sublattice represents a set of sites in a supercell that all have
the same domain/site space. More rigourously it represents a substructure
of the random structure supercell being sampled in a Monte Carlo
simulation.
"""

__author__ = "Luis Barroso-Luque"

import itertools
from dataclasses import dataclass, field

import numpy as np
from monty.json import MSONable
from pymatgen.core import Composition

from smol.cofe.space.domain import SiteSpace, Vacancy


@dataclass
class Sublattice(MSONable):
    """Sublattice class.

    A Sublattice is used to represent a subset of supercell sites that have
    the same site space. Rigorously it represents a set of sites in a
    "substructure" of the total structure.

    Attributes:
     site_space (SiteSpace):
        SiteSpace with the allowed species and their random
        state composition.
     sites (ndarray):
        array of site indices for all sites in sublattice
     active_sites (ndarray):
        array of site indices for all unrestricted sites in the sublattice.
     encoding (ndarray):
        array of species encoding in integer indices. By default,
        will be initialized as range(len(site_space)). Might be different
        if a sub-lattice was created from the split of another sub-lattice.
    """

    site_space: SiteSpace
    sites: np.array
    active_sites: np.array = field(init=False)
    encoding: np.array = field(init=False)

    def __post_init__(self):
        """Copy sites into active_sites, and initial setup."""
        self.sites = np.unique(self.sites)
        self.active_sites = self.sites.copy()
        if len(self.site_space) <= 1:
            # A single-species sub-lattice should not be active at all.
            self.restrict_sites(self.sites)

        self.encoding = np.arange(len(self.site_space), dtype=int)

    @property
    def is_active(self):
        """Whether sub-lattice has active sites."""
        return len(self.active_sites) > 0

    @property
    def species(self):
        """Get allowed species for sites in sublattice."""
        return tuple(self.site_space.keys())

    @property
    def restricted_sites(self):
        """Get restricted sites for species."""
        return np.setdiff1d(self.sites, self.active_sites)

    def restrict_sites(self, sites):
        """Restricts (freezes) the given sites.

        Once a site is restricted, no Metropolis step can be proposed
        with it, including flipping, swapping, etc.

        Args:
            sites (Sequence):
                indices of sites in the occupancy string to restrict.
        """
        self.active_sites = np.array([i for i in self.active_sites if i not in sites])

    def reset_restricted_sites(self):
        """Reset all restricted sites to active."""
        # Single species sub-lattice can never be active.
        if len(self.site_space) > 1:
            self.active_sites = self.sites.copy()

    def split_by_species(self, occu, species_in_partitions):
        """Split a sub-lattice into multiple by specie.

        An example use case might be simulating topotactic Li extraction
        and insertion, where we want to consider Li/Vac, TM and O as
        different sub-lattices that can not be mixed by swapping.
        Args:
            occu (np.ndarray[int]):
                An occupancy array to reference with.
            species_in_partitions (List[List[int|Species|Vacancy|Element|str]]):
                Each sub-list contains a few species or encodings of species in
                the site space to be grouped as a new sub-lattice, namely,
                sites with occu[sites] == specie in the sub-list, will be
                used to initialize a new sub-lattice.
                Sub-lists will be pre-sorted to ascending order.
        Returns:
            List of split sub-lattices:
                List[Sublattice]
        """
        part_sublattices = []
        # Codes given
        if all(
            isinstance(sp, (int, np.int32, np.int64))
            for sp in itertools.chain(*species_in_partitions)
        ):
            codes_in_partitions = species_in_partitions
        # Species or species strings given.
        else:
            # Treat vacancies more carefully. One sub-lattice can only
            # have one vacancy species.
            def get_index(sp, species):
                if isinstance(sp, Vacancy):
                    for i, sp2 in enumerate(species):
                        if isinstance(sp2, Vacancy):
                            return i
                return species.index(sp)

            codes_in_partitions = [
                [self.encoding[get_index(sp, self.species)] for sp in partition]
                for partition in species_in_partitions
            ]

        for species_codes in codes_in_partitions:
            part_comp = {}
            part_sites = []
            part_actives = []
            # Because site space species were sorted.
            part_codes = sorted(species_codes)
            for code in part_codes:
                sp_id = np.where(self.encoding == code)[0][0]
                sp = self.species[sp_id]
                part_comp[sp] = self.site_space[sp]
                part_sites.extend(self.sites[occu[self.sites] == code].tolist())
                part_actives.extend(
                    self.active_sites[occu[self.active_sites] == code].tolist()
                )
            # Re-weighting partitioned site-space
            part_n = sum(list(part_comp.values()))
            part_comp = {
                sp: part_comp[sp] / part_n
                for sp in part_comp
                if not isinstance(sp, Vacancy)
            }
            part_comp = Composition(part_comp)
            part_space = SiteSpace(part_comp)
            part_sites = np.array(part_sites, dtype=int)
            part_actives = np.array(part_actives, dtype=int)
            part_codes = np.array(part_codes, dtype=int)
            part_sublatt = Sublattice(part_space, part_sites)
            part_sublatt.active_sites = part_actives
            part_sublatt.encoding = part_codes
            if len(part_codes) == 1:
                part_sublatt.restrict_sites(part_sublatt.sites)
            part_sublattices.append(part_sublatt)
        return part_sublattices

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        sl_d = {
            "site_space": self.site_space.as_dict(),
            "sites": self.sites.tolist(),
            "encoding": self.encoding.tolist(),
            "active_sites": self.active_sites.tolist(),
        }
        return sl_d

    @classmethod
    def from_dict(cls, d):
        """Instantiate a sublattice from dict representation.

        Returns:
            Sublattice
        """
        sublattice = cls(
            SiteSpace.from_dict(d["site_space"]), sites=np.array(d["sites"], dtype=int)
        )
        sublattice.active_sites = np.array(d["active_sites"], dtype=int)
        sublattice.encoding = np.array(d["encoding"], dtype=int)
        return sublattice



"""Implementation of an Orbit.

A set of symmetrically equivalent (with respect to the given random structure
symmetry) clusters.
"""

import operator
from functools import reduce
from itertools import chain, combinations, product

import numpy as np
from monty.json import MSONable
from pymatgen.core import Lattice
from pymatgen.core.operations import SymmOp
from pymatgen.util.coord import coord_list_mapping, is_coord_subset

from smol.exceptions import SYMMETRY_ERROR_MESSAGE, SymmetryError

from .basis import DiscreteBasis, basis_factory
from .cluster import Cluster
from .constants import SITE_TOL

__author__ = "Luis Barroso-Luque, William Davidson Richard"


class Orbit(MSONable):
    """Orbit set of symmetrically equivalent clusters.

    An Orbit represents a set of clusters that are symmetrically equivalent
    in the random structure. The class also includes the possible orderings on
    the clusters in the orbit. The different orderings represent the single
    site function indices to generate all possible orbit functions (correlation
    functions) for the given orbit.

    An orbit usually includes translational and structure symmetry of the
    underlying lattice. But this is not a hard requirement, and any set of
    symmetry operations can be passed to the constructor. Regardless, an orbit
    should at a minimum have translational symmetry.

    You probably never need to instantiate this class directly. Look at
    ClusterSubspace to create orbits and clusters necessary for a CE.

    Attributes:
        bits (list of list):
            List of lists describing the possible non-constant site function
            indices at each site of a cluster in the orbit.
        site_bases (list of DiscreteBasis):
            List of the SiteBasis for each site.
        structure_symops (list of Symmops):
            List of underlying structure symmetry operations.
        lattice (Lattice):
            Underlying Structure's lattice.
    """

    def __init__(self, cluster_coords, lattice, bits, site_bases, structure_symops):
        """Initialize an Orbit.

        Args:
            cluster_coords (list or ndarray):
                list of frac coords for the sites in base cluster.
            lattice (pymatgen.Lattice):
                a Lattice object for the given sites
            bits (list of list):
                list describing the possible site function indices for
                each site in cluster. Should be the number of possible
                occupancies minus one. For a 3 site cluster, each of which
                can have one of Li, TM, or Vac, bits are [[0, 1], [0, 1], [0, 1]]
                This is ensures the expansion is not "over-complete" by
                implicitly enforcing that all sites have a site basis function
                phi_0 = 1.
            site_bases (list of DiscreteBasis):
                list of SiteBasis objects for each site in the given sites.
            structure_symops (list of SymmOp):
                list of symmetry operations for the base structure
        """
        if len(cluster_coords) != len(bits):
            raise AttributeError(
                f"Number of sites {len(cluster_coords)} must be equal to number of "
                f"bits {len(bits)}"
            )
        if len(cluster_coords) != len(site_bases):
            raise AttributeError(
                f"Number of sites {len(cluster_coords)} must be equal to number of "
                f"site bases {len(site_bases)}"
            )

        self.bits = bits
        self.site_bases = site_bases
        self.structure_symops = structure_symops

        # ids should be assigned using the assign_id method externally
        self.id = None  # id identifying orbit amongst all other orbits only
        self.bit_id = None  # id for first bit combo in this orbit
        # considering all the bit combos in all orbits.

        # lazy generation of properties
        self._equiv = None
        self._symops = None
        self._permutations = None
        self._bit_combos = None
        self._basis_arrs = None
        self._corr_tensors = None
        self._flat_corr_tensors = None

        # Create basecluster
        self.base_cluster = Cluster(
            [site_basis.site_space for site_basis in site_bases],
            cluster_coords,
            lattice,
        )

    @property
    def basis_type(self):
        """Return the name of basis set used."""
        return self.site_bases[0].flavor

    @property
    def basis_orthogonal(self):
        """Test if the Orbit bases are orthogonal."""
        return all(basis.is_orthogonal for basis in self.site_bases)

    @property
    def basis_orthonormal(self):
        """Test if the orbit bases are orthonormal."""
        return all(basis.is_orthonormal for basis in self.site_bases)

    @property
    def multiplicity(self):
        """Get number of clusters in orbit.

        Number of clusters in the given sites of the lattice object that are in
        the orbit.
        """
        return len(self.clusters)

    @property
    def bit_combos(self):
        """Get tuple representing the multi-index for site function ordering.

        tuple of ndarrays, each array is a set of symmetrically equivalent bit
        orderings represented by each row. Bit combos represent non-constant site
        function orderings, i.e. contracted multi-indices
        """
        if self._bit_combos is None:
            # get all the bit symmetry operations
            all_combos = []
            for bit_combo in product(*self.bits):
                if not any(np.array_equal(bit_combo, bc) for bc in chain(*all_combos)):
                    bit_combo = np.array(bit_combo, dtype=np.int_)
                    new_bits = np.unique(bit_combo[self.cluster_permutations], axis=0)
                    all_combos.append(new_bits)
            self._bit_combos = tuple(all_combos)
        return self._bit_combos

    @property
    def site_spaces(self):
        """Get the site spaces for the site basis associated with each site."""
        return [site_basis.site_space for site_basis in self.site_bases]

    @property
    def bit_combo_multiplicities(self):
        """Get the multiplicities of the symmetrically distinct bit ordering."""
        return [bcombo.shape[0] for bcombo in self.bit_combos]

    @property
    def clusters(self):
        """Get symmetrically equivalent clusters."""
        if self._equiv:
            return self._equiv
        equiv = [self.base_cluster]
        for symop in self.structure_symops:
            new_coords = symop.operate_multi(self.base_cluster.frac_coords)
            cluster = Cluster(
                [site_basis.site_space for site_basis in self.site_bases],
                new_coords,
                self.base_cluster.lattice,
            )
            if cluster not in equiv:
                equiv.append(cluster)
        self._equiv = equiv
        if len(equiv) * len(self.cluster_symops) != len(self.structure_symops):
            self._equiv = None  # Unset this
            raise SymmetryError(SYMMETRY_ERROR_MESSAGE)
        return equiv

    @property
    def cluster_symops(self):
        """Get symmetry operations that map a cluster to its periodic image.

        Each element is a pymatgen.core.operations.Symop.
        """
        if self._symops is None:
            self._gen_cluster_symops()
        return self._symops

    @property
    def cluster_permutations(self):
        """Get the symmetrical site permutations that map a cluster to itself.

        A permutation is a mapping such that for a give symop in cluster_symops
        Symop.operate(sites) = sites[mapping] (after translation back to unit
        cell)
        """
        if self._permutations is None:
            self._gen_cluster_symops()
        return self._permutations

    @property
    def basis_arrays(self):
        """Get a tuple of all site function arrays for each site in orbit."""
        if self._basis_arrs is None:
            self._basis_arrs = tuple(sb.function_array for sb in self.site_bases)
        return self._basis_arrs

    @property
    def correlation_tensors(self):
        """Get the array of correlation functions for all possible configs.

        Array of stacked correlation arrays for each symetrically distinct
        set of bit combos.

        The correlations array is a multidimensional array with each dimension
        corresponding to each site space.

        First dimension is for bit_combos, and the remaining dimensions correspond
        to site spaces.

        i.e. correlation_tensors[0, 1, 0, 2] gives the value of the
        correlation function for bit_combo 0 evaluated for a cluster with
        occupancy [1, 0, 2]
        """
        if self._corr_tensors is None:
            corr_tensors = np.zeros(
                (len(self.bit_combos), *(basis.shape[1] for basis in self.basis_arrays))
            )

            for i, combos in enumerate(self.bit_combos):
                for bits in combos:
                    corr_tensors[i] += reduce(
                        lambda a, b: np.tensordot(a, b, axes=0),
                        (self.basis_arrays[i][b] for i, b in enumerate(bits)),
                    )
                corr_tensors[i] /= len(combos)
            self._flat_corr_tensors = None  # reset
            self._corr_tensors = corr_tensors
        return self._corr_tensors

    @property
    def flat_correlation_tensors(self):
        """Get correlation_tensors flattened to 2D for fast cython."""
        if self._flat_corr_tensors is None:
            self._flat_corr_tensors = np.ascontiguousarray(
                np.reshape(
                    self.correlation_tensors,
                    (
                        self.correlation_tensors.shape[0],
                        np.prod(self.correlation_tensors.shape[1:]),
                    ),
                    order="C",
                )
            )
        return self._flat_corr_tensors

    @property
    def flat_tensor_indices(self):
        """Index multipliers to read data easier from flat corr tensors."""
        indices = np.cumprod(np.append(self.correlation_tensors.shape[2:], 1)[::-1])[
            ::-1
        ]
        return np.ascontiguousarray(indices, dtype=int)

    @property
    def rotation_array(self):
        """Get the rotation array.

        The rotation array is of size len(bit combos) x len(bit combos)
        """
        rot_array = np.empty(2 * (len(self._bit_combos),))
        for (i, j), (bcombos_i, bcombos_j) in zip(
            product(range(len(self._bit_combos)), repeat=2),
            product(self._bit_combos, repeat=2),
        ):
            rot_array[i, j] = sum(
                reduce(
                    operator.mul,
                    (
                        np.dot(
                            self.site_bases[k].rotation_array.T
                            @ self.basis_arrays[k][bj],  # noqa
                            self.site_bases[k].measure_vector
                            * self.basis_arrays[k][bi],  # noqa
                        )
                        for k, (bi, bj) in enumerate(zip(bcombo_i, bcombo_j))
                    ),
                )
                for bcombo_i, bcombo_j in product(bcombos_i, bcombos_j)
            ) / len(bcombos_i)
            # \ (len(bcombos_i) * len(bcombos_j))**0.5 is unitary
        return rot_array

    def remove_bit_combo(self, bits):  # seems like this is no longer used?
        """Remove bit_combos from orbit.

        Only a single set of bits in the bit combo (symmetrically equivalent bit
        orderings) needs to be passed.
        """
        bit_combos = []

        for bit_combo in self.bit_combos:
            if not any(np.array_equal(bits, b) for b in bit_combo):
                bit_combos.append(bit_combo)

        if not bit_combos:
            raise RuntimeError(
                "All bit_combos have been removed from orbit with id " f"{self.id}"
            )

        self._bit_combos = tuple(bit_combos)
        self.reset_bases()

    def remove_bit_combos_by_inds(self, inds):
        """Remove bit combos by their indices in the bit_combo list."""
        if max(inds) > len(self.bit_combos) - 1:
            raise ValueError(
                f"Some indices {inds} out of range for total "
                f"{len(self._bit_combos)} bit combos"
            )

        self._bit_combos = tuple(
            b_c for i, b_c in enumerate(self._bit_combos) if i not in inds
        )

        if not self.bit_combos:
            raise RuntimeError(
                "All bit_combos have been removed from orbit with id " f"{self.id}"
            )
        self.reset_bases()

    def transform_site_bases(self, basis_name, orthonormal=False):
        """Transform the Orbits site bases into new basis set.

        Args:
            basis_name (str):
                name of new basis for all site bases
            orthonormal (bool):
                option to orthonormalize all new site bases
        """
        new_bases = []
        for basis in self.site_bases:
            new_basis = basis_factory(basis_name, basis.site_space)
            if orthonormal:
                new_basis.orthonormalize()
            new_bases.append(new_basis)

        self.site_bases = tuple(new_bases)
        self.reset_bases()

    def is_sub_orbit(self, orbit):
        """Check if given orbits clusters are subclusters.

        Note this does not consider bit_combos

        Args:
            orbit (Orbit):
                Orbit object to check if its clusters are subclusters
        Returns:
            bool: True if the clusters of given orbit are subclusters.
        """
        if len(self.base_cluster) <= len(orbit.base_cluster):
            return False

        if not np.all(sp in self.site_spaces for sp in orbit.site_spaces):
            return False

        match = any(
            Cluster(
                [site_basis.site_space for site_basis in self.site_bases],
                self.base_cluster.frac_coords[inds, :],
                self.base_cluster.lattice,
            )
            in orbit.clusters
            for inds in combinations(
                range(len(self.base_cluster)), len(orbit.base_cluster)
            )
        )

        return match

    def sub_orbit_mappings(self, orbit):
        """Return a mapping of the sites in the orbit to a sub orbit.

        If the given orbit is not a sub-orbit will return an empty list.
        Note this works for mapping between sites, sites spaces, and basis
        functions associated with each site.

        Args:
            orbit (Orbit):
                A sub orbit to return mapping of sites
        Returns:
            list: of indices sucht that
                self.base_cluster.sites[indices] = orbit.base_cluster.sites
        """
        indsets = np.array(
            list(
                combinations(
                    (
                        i
                        for i, space in enumerate(self.site_spaces)
                        if space in orbit.site_spaces
                    ),
                    len(orbit.site_spaces),
                )
            ),
            dtype=int,
        )

        mappings = []
        for cluster in self.clusters:
            for inds in indsets:
                # take the centroid of subset of sites, not all cluster sites
                centroid = np.average(cluster.frac_coords[inds], axis=0)
                recenter = np.round(centroid - orbit.base_cluster.centroid)
                c_sites = orbit.base_cluster.frac_coords + recenter
                if is_coord_subset(c_sites, cluster.frac_coords):
                    mappings.append(
                        coord_list_mapping(c_sites, cluster.frac_coords, atol=SITE_TOL)
                    )

        if len(mappings) == 0 and self.is_sub_orbit(orbit):
            raise RuntimeError(
                "The given orbit is a suborbit, but no site mappings were "
                "found!\n Something is very wrong here!"
            )
        return np.unique(mappings, axis=0)

    def _gen_cluster_symops(self):
        """Generate the cluster SymOps and decoration permutations."""
        symops = []
        permutations = []
        for symop in self.structure_symops:
            new_sites = symop.operate_multi(self.base_cluster.frac_coords)
            cluster = Cluster(
                [site_basis.site_space for site_basis in self.site_bases],
                new_sites,
                self.base_cluster.lattice,
            )
            if cluster == self.base_cluster:
                recenter = np.round(self.base_cluster.centroid - cluster.centroid)
                c_sites = cluster.frac_coords + recenter
                mapping = coord_list_mapping(
                    self.base_cluster.frac_coords, c_sites, atol=SITE_TOL
                )
                symops.append(symop)
                permutations.append(mapping)

        self._permutations = np.unique(permutations, axis=0)
        self._symops = tuple(symops)

        if len(self._symops) * self.multiplicity != len(self.structure_symops):
            raise SymmetryError(SYMMETRY_ERROR_MESSAGE)

    def reset_bases(self):
        """Reset cached basis function array and correlation tensors."""
        self._basis_arrs = None
        self._corr_tensors = None
        self._flat_corr_tensors = None

    def assign_ids(self, orbit_id, orbit_bit_id, start_cluster_id):
        """Assign unique orbit and cluster ID's.

        This should be called iteratively for a list of orbits to get a proper
        set of unique ID's for the orbits.

        Args:
            orbit_id (int): orbit id
            orbit_bit_id (int): starting bit ordering id
            start_cluster_id (int): starting cluster id

        Returns:
            (int, int, int):
            next orbit id, next bit ordering id, next cluster id
        """
        self.id = orbit_id
        self.bit_id = orbit_bit_id
        c_id = start_cluster_id
        for cluster in self.clusters:
            c_id = cluster.assign_ids(c_id)
        return orbit_id + 1, orbit_bit_id + len(self.bit_combos), c_id

    def __len__(self):
        """Get total number of orbit basis functions.

        The number of symmetrically distinct bit orderings.
        """
        return len(self.bit_combos)

    def __eq__(self, other):
        """Check equality of orbits (only compares crystallographic equivalence)."""
        # when performing orbit in list, this ordering stops the
        # equivalent structures from generating
        return self.base_cluster in other.clusters

    def __contains__(self, cluster):
        """Check if a cluster is included in orbit."""
        return cluster in self.clusters

    def __str__(self):
        """Pretty strings for pretty things."""
        outs = [
            f"Orbit  {self.id}",
            f"    Multiplicity : {self.multiplicity:<4}",
            f"   No. functions : {len(self):<4}",
            f"No. symmetry ops : {len(self.cluster_symops):<4}",
            f"Function ids : {list(range(self.bit_id, self.bit_id + len(self)))}",
            "Base Cluster : ",
            "  | " + "\n  | ".join(str(self.base_cluster).split("\n")),
        ]
        return "\n".join(outs)

    def __repr__(self):
        """Get Orbit representation."""
        outs = [
            f"Orbit {self.id}",
            f"Multiplicity: {self.multiplicity:<4}  No. functions: {len(self):<4}  "
            f"No. symmetry ops: {len(self.cluster_symops):<4}",
            f"Function ids: {list(range(self.bit_id, self.bit_id + len(self)))}",
            "Base Cluster:\n  " + "\n  ".join(repr(self.base_cluster).split("\n")[1:]),
        ]
        return "\n".join(outs)

    @classmethod
    def from_dict(cls, d):
        """Create Orbit from serialized MSONable dict."""
        structure_symops = [SymmOp.from_dict(so_d) for so_d in d["structure_symops"]]
        site_bases = [DiscreteBasis.from_dict(sd) for sd in d["site_bases"]]
        orb = cls(
            d["sites"],
            Lattice.from_dict(d["lattice"]),
            d["bits"],
            site_bases,
            structure_symops,
        )

        orb._bit_combos = (
            tuple(np.array(c, dtype=int) for c in d["_bit_combos"])
            if "_bit_combos" in d
            else None
        )
        # This is to ensure that, after removing some bit_combos, an orbit
        # can still be correctly recorded and reloaded.
        return orb

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        orb_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "sites": self.base_cluster.frac_coords.tolist(),
            "lattice": self.base_cluster.lattice.as_dict(),
            "bits": self.bits,
            "site_bases": [sb.as_dict() for sb in self.site_bases],
            "structure_symops": [so.as_dict() for so in self.structure_symops],
            "_bit_combos": tuple(c.tolist() for c in self.bit_combos),
        }
        return orb_d



"""
Implementation of Relevant classes for MCMC Sampling.

Samplers should derive from ABC Sampler classes whenever possible.

MCMCUshers are used to propose different kinds of MC steps for specific
simulations.

Samples are saved in a Sample Container.
"""

from smol.moca.sampler.container import SampleContainer

__all__ = ["SampleContainer"]



"""
Class implementations to create Cluster Expansions.

The cofe (pronounced coffee) package contains all the necessary classes and
functions to define and fit cluster expansions for crystalline materials.
"""

from smol.cofe.wrangling.wrangler import StructureWrangler

from .expansion import ClusterExpansion, RegressionData
from .space.clusterspace import ClusterSubspace, PottsSubspace

__all__ = [
    "ClusterSubspace",
    "PottsSubspace",
    "StructureWrangler",
    "ClusterExpansion",
    "RegressionData",
]



import numpy as np
import pytest
from pymatgen.core import Composition

from smol.cofe.space.domain import SiteSpace
from smol.moca.sampler.mcusher import Flip, Swap
from smol.moca.sublattice import Sublattice
from tests.utils import gen_random_occupancy

mcmcusher_classes = [Flip, Swap]
num_sites = 100


@pytest.fixture
def all_sublattices(rng):
    # generate two tests sublattices
    sites = np.arange(num_sites)
    sites1 = rng.choice(sites, size=num_sites // 3)
    sites2 = rng.choice(np.setdiff1d(sites, sites1), size=num_sites // 4)
    sites3 = np.setdiff1d(sites, np.concatenate((sites1, sites2)))
    site_space1 = SiteSpace(Composition({"A": 0.1, "B": 0.4, "C": 0.3, "D": 0.2}))
    site_space2 = SiteSpace(Composition({"A": 0.1, "B": 0.4, "E": 0.5}))
    site_space3 = SiteSpace(Composition({"G": 1}))
    active_sublattices = [
        Sublattice(site_space1, sites1),
        Sublattice(site_space2, sites2),
    ]
    inactive_sublattices = [Sublattice(site_space3, sites3)]
    return active_sublattices, inactive_sublattices


@pytest.fixture
def rand_occu(all_sublattices):
    # generate a random occupancy according to the sublattices
    occu = gen_random_occupancy(all_sublattices[0] + all_sublattices[1])
    return occu, all_sublattices[1][0].sites  # return indices of fixed sites


@pytest.fixture(params=mcmcusher_classes)
def mcmcusher(request, all_sublattices):
    # instantiate mcmcushers to test
    return request.param(all_sublattices[0] + all_sublattices[1])


def test_bad_propabilities(mcmcusher):
    with pytest.raises(ValueError):
        mcmcusher.sublattice_probabilities = [0.6, 0.1]
    with pytest.raises(AttributeError):
        mcmcusher.sublattice_probabilities = [0.5, 0.2, 0.3]


def test_propose_step(mcmcusher, rand_occu):
    occu, fixed_sites = rand_occu
    iterations = 50000
    # test with 50/50 probability
    flipped_sites = []
    count1, count2 = 0, 0
    total = 0
    for i in range(iterations):
        step = mcmcusher.propose_step(occu)
        for flip in step:
            assert flip[1] != occu[flip[0]]
            if flip[0] in mcmcusher.active_sublattices[0].active_sites:
                count1 += 1
                assert flip[1] in mcmcusher.active_sublattices[0].encoding
            elif flip[0] in mcmcusher.active_sublattices[1].active_sites:
                count2 += 1
                assert flip[1] in mcmcusher.active_sublattices[1].encoding
            else:
                raise RuntimeError(
                    "Something went wrong in proposing"
                    f"a step site proposed in {step} is"
                    " not in any of the allowed sites"
                )
            total += 1
            flipped_sites.append(flip[0])

    # check probabilities seem sound
    assert count1 / total == pytest.approx(0.5, abs=1e-2)
    assert count2 / total == pytest.approx(0.5, abs=1e-2)

    # check that every site was flipped at least once
    assert all(
        i in flipped_sites for i in np.setdiff1d(np.arange(num_sites), fixed_sites)
    )

    # make sure fixed sites remain the same
    assert all(i not in fixed_sites for i in flipped_sites)

    # Now check with a sublattice bias
    mcmcusher.sublattice_probabilities = [0.8, 0.2]
    flipped_sites = []
    count1, count2 = 0, 0
    total = 0
    for i in range(iterations):
        step = mcmcusher.propose_step(occu)
        for flip in step:
            if flip[0] in mcmcusher.active_sublattices[0].active_sites:
                count1 += 1
                assert flip[1] in mcmcusher.active_sublattices[0].encoding
            elif flip[0] in mcmcusher.active_sublattices[1].sites:
                count2 += 1
                assert flip[1] in mcmcusher.active_sublattices[1].encoding
            else:
                raise RuntimeError(
                    "Something went wrong in proposing"
                    f"a step site proposed in {step} is"
                    " not in any of the allowed sites"
                )
            total += 1
            flipped_sites.append(flip[0])
    assert count1 / total == pytest.approx(0.8, abs=1e-2)
    assert count2 / total == pytest.approx(0.2, abs=1e-2)



"""Input and output functions.

Some convenience functions to save and load CE and MC workflows in a "standard"
way.

Can also at some point add input and output for other CE code packages to make
it easier to translate work.
"""

__author__ = "Luis Barroso-Luque"

import json

from monty.json import MontyDecoder, MontyEncoder, MSONable


def save_work(file_path, *msonables):
    """Save msonable classes used in a CE and/or MC workflow.

    Save a set of unique MSONable objects used in a workflow for a cluster
    expansion and/or monte carlo run as a json dictionary.

    Only one of each class type can be saved per file.

    Args:
        file_path (str):
            file path
        *msonables (monty.MSONable):
            MSONable child classes.
    """
    work_d = {}
    for msonable in msonables:
        if not isinstance(msonable, MSONable):
            raise AttributeError(
                "Attempting to save an object which is not " f"MSONable: {msonable}"
            )
        work_d[msonable.__class__.__name__] = msonable

    with open(file_path, "w", encoding="utf-8") as fpath:
        json.dump(work_d, fpath, cls=MontyEncoder)


def load_work(file_path):
    """Load a dictionary and instantiate the MSONable objects.

    Args:
        file_path (str):

    Returns: Dictionary with smol objects
        dict
    """
    with open(file_path, encoding="utf-8") as fpath:
        work_d = json.load(fpath, cls=MontyDecoder)

    return work_d



#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False
#cython: initializedcheck=False
#cython: cdivision=True

"""
General cython utility functions to calculate correlation vectors and local
changes in correlation vectors a tad bit faster than pure python.
"""

__author__ = "Luis Barroso-Luque, William D. Richards"

import numpy as np
cimport numpy as np


cpdef corr_from_occupancy(const long[::1] occu,
                          const int num_corr_functions,
                          list orbit_list):
    """Computes the correlation vector for a given encoded occupancy string.

    Args:
        occu (ndarray):
            encoded occupancy vector
        num_corr_functions (int):
            total number of bit orderings in expansion.
        orbit_list:
            Information of all orbits.
            (orbit id, flat tensor index array, flat correlation tensor,
             site indices of clusters)

    Returns: array
        correlation vector for given occupancy
    """
    cdef int i, j, n, m, I, J, M, index
    cdef double p
    cdef const long[:, ::1] indices
    cdef const long[::1] tensor_indices
    cdef const double[:, ::1] corr_tensors
    out = np.zeros(num_corr_functions)
    cdef double[:] o_view = out
    o_view[0] = 1  # empty cluster

    for n, tensor_indices, corr_tensors, indices in orbit_list:
        M = corr_tensors.shape[0]  # index of bit combos
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        for m in range(M):
            p = 0
            for i in range(I):
                index = 0
                for j in range(J):
                    index += tensor_indices[j] * occu[indices[i, j]]
                p += corr_tensors[m, index]
            o_view[n] = p / I
            n += 1
    return out


cpdef delta_corr_single_flip(const long[::1] occu_f,
                             const long[::1] occu_i,
                             const int num_corr_functions,
                             list site_orbit_list):
    """Computes the correlation difference between two occupancy vectors.

    Args:
        occu_f (ndarray):
            encoded occupancy vector with flip
        occu_i (ndarray):
            encoded occupancy vector without flip
        num_corr_functions (int):
            total number of bit orderings in expansion.
        site_orbit_list:
            Information of all orbits that include the flip site.
            List of tuples each with
            (orbit id, cluster ratio, flat tensor index array,
             flat correlation tensor, site indices of clusters)


    Returns:
        ndarray: correlation vector difference
    """
    cdef int i, j, n, m, I, J, M, ind_i, ind_f
    cdef double p, ratio
    cdef const long[:, ::1] indices
    cdef const long[::1] tensor_indices
    cdef const double[:, ::1] corr_tensors
    out = np.zeros(num_corr_functions)
    cdef double[::1] o_view = out

    for n, ratio, tensor_indices, corr_tensors, indices in site_orbit_list:
        M = corr_tensors.shape[0] # index of bit combos
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        for m in range(M):
            p = 0
            for i in range(I):
                ind_i, ind_f = 0, 0
                for j in range(J):
                    ind_i += tensor_indices[j] * occu_i[indices[i, j]]
                    ind_f += tensor_indices[j] * occu_f[indices[i, j]]
                p += (corr_tensors[m, ind_f] - corr_tensors[m, ind_i])
            o_view[n] = p / ratio / I
            n += 1
    return out


cpdef indicator_delta_corr_single_flip(const long[::1] occu_f,
                                       const long[::1] occu_i,
                                       const int num_corr_functions,
                                       list site_orbit_list):
    """Local change in indicator basis correlation vector from single flip.

    Args:
        occu_f (ndarray):
            encoded occupancy vector with flip
        occu_i (ndarray):
            encoded occupancy vector without flip
        num_corr_functions (int):
            total number of bit orderings in expansion.
        site_orbit_list:
            Information of all orbits that include the flip site.
            List of tuples each with
            (orbit id, cluster ratio, bit_combos,
             bit_combo_indices site indices, bases array)

    Returns:
        ndarray: correlation vector difference
    """
    cdef int i, j, k, n, m, I, K, M
    cdef bint ok
    cdef const long[:, ::1] bit_combos, indices
    cdef const long[::1] bit_indices
    out = np.zeros(num_corr_functions)
    cdef double[::1] o_view = out
    cdef double r, o

    for n, r, bit_combos, bit_indices, _, indices in site_orbit_list:
        M = bit_indices.shape[0] # index of bit combos
        I = indices.shape[0] # cluster index
        K = indices.shape[1] # index within cluster
        for m in range(M - 1):
            o = 0
            for i in range(I):
                for j in range(bit_indices[m], bit_indices[m + 1]):
                    ok = True
                    for k in range(K):
                        if occu_f[indices[i, k]] != bit_combos[j, k]:
                            ok = False
                            break
                    if ok:
                        o += 1

                    ok = True
                    for k in range(K):
                        if occu_i[indices[i, k]] != bit_combos[j, k]:
                            ok = False
                            break
                    if ok:
                        o -= 1

            o_view[n] = o / r / (I * (bit_indices[m + 1] - bit_indices[m]))
            n += 1
    return out


cpdef delta_ewald_single_flip(const long[::1] occu_f,
                              const long[::1] occu_i,
                              const double[:, ::1] ewald_matrix,
                              const long[:, ::1] ewald_indices,
                              const int site_ind):
    """Compute the change in electrostatic interaction energy from a flip.

    Args:
        occu_f (ndarray):
            encoded occupancy vector with flip
        occu_i (ndarray):
            encoded occupancy vector without flip
        ewald_matrix (ndarray):
            Ewald matrix for electrostatic interactions
        ewald_indices (ndarray):
            2D array of indices corresponding to a specific site occupation
            in the ewald matrix
        site_ind (int):
            site index for site being flipped

    Returns:
        float: electrostatic interaction energy difference
    """
    cdef int i, j, k, add, sub
    cdef bint ok
    cdef double out = 0
    cdef double out_k

    # values of -1 are vacancies and hence don't have ewald indices
    add = ewald_indices[site_ind, occu_f[site_ind]]
    sub = ewald_indices[site_ind, occu_i[site_ind]]

    for k in range(occu_f.shape[0]):
        i = ewald_indices[k, occu_f[k]]
        out_k = 0
        if i != -1 and add != -1:
            if i != add:
                out_k = out_k + 2 * ewald_matrix[i, add]
            else:
                out_k = out_k + ewald_matrix[i, add]

        j = ewald_indices[k, occu_i[k]]
        if j != -1 and sub != -1:
            if j != sub:
                out_k = out_k - 2 * ewald_matrix[j, sub]
            else:
                out_k = out_k - ewald_matrix[j, sub]
        out += out_k
    return out


cpdef corr_from_occupancy_low_mem(const long[::1] occu,
                                  const int num_corr_functions,
                                  list orbit_list):
    """Computes the correlation vector for a given encoded occupancy string.

    Lower memory usage but slightly slower and worse scaling

    Args:
        occu (ndarray):
            encoded occupancy vector
        num_corr_functions (int):
            total number of bit orderings in expansion.
        orbit_list:
            Information of all orbits.
            (orbit id, bit_combos, bit_combo_indices site indices, bases array)

    Returns: array
        correlation vector for given occupancy
    """
    cdef int i, j, k, n, m, I, K, M
    cdef double p, pi
    cdef const long[:, ::1] indices, bit_combos
    cdef const long[::1] bit_indices
    cdef const double[:, :, ::1] bases
    out = np.zeros(num_corr_functions)
    cdef double[:] o_view = out
    o_view[0] = 1  # empty cluster

    for n, bit_combos, bit_indices, bases, indices in orbit_list:
        M = bit_indices.shape[0] # index of bit combos
        I = indices.shape[0] # cluster index
        K = indices.shape[1] # index within cluster
        for m in range(M - 1):
            p = 0
            for i in range(I):
                for j in range(bit_indices[m], bit_indices[m + 1]):
                    pi = 1
                    for k in range(K):
                        pi *= bases[k, bit_combos[j, k], occu[indices[i, k]]]
                    p += pi
            o_view[n] = p / (I * (bit_indices[m + 1] - bit_indices[m]))
            n += 1
    return out


cpdef delta_corr_single_flip_lm(const long[::1] occu_f,
                                     const long[::1] occu_i,
                                     const int num_corr_functions,
                                     list site_orbit_list):
    """Computes the correlation difference between two occupancy vectors.

    Lower memory usage but slightly slower and worse scaling

    Args:
        occu_f (ndarray):
            encoded occupancy vector with flip
        occu_i (ndarray):
            encoded occupancy vector without flip
        num_corr_functions (int):
            total number of bit orderings in expansion.
        site_orbit_list:
            Information of all orbits that include the flip site.
            List of tuples each with
            (orbit id, clustar ratio, bit_combos,
            bit_combo_indices site indices, bases array)


    Returns:
        ndarray: correlation vector difference
    """
    cdef int i, j, k, n, m, I, K, M
    cdef double p, pi, pf, r
    cdef const long[:, ::1] indices, bit_combos
    cdef const long[::1] bit_indices
    cdef const double[:, :, ::1] bases
    out = np.zeros(num_corr_functions)
    cdef double[::1] o_view = out

    for n, r, bit_combos, bit_indices, bases, indices in site_orbit_list:
        M = bit_indices.shape[0] # index of bit combos
        I = indices.shape[0] # cluster index
        K = indices.shape[1] # index within cluster
        for m in range(M - 1):
            p = 0
            for i in range(I):
                for j in range(bit_indices[m], bit_indices[m + 1]):
                    pf = 1
                    pi = 1
                    for k in range(K):
                        pf *= bases[k, bit_combos[j, k], occu_f[indices[i, k]]]
                        pi *= bases[k, bit_combos[j, k], occu_i[indices[i, k]]]
                    p += (pf - pi)
            o_view[n] = p / r / (I * (bit_indices[m + 1] - bit_indices[m]))
            n += 1
    return out



from copy import deepcopy

import numpy as np
import numpy.testing as npt
import pytest

from smol.cofe import ClusterExpansion, RegressionData
from smol.moca import (
    CanonicalEnsemble,
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
    SemiGrandEnsemble,
)
from tests.utils import assert_msonable, gen_random_occupancy

ensembles = [CanonicalEnsemble, SemiGrandEnsemble]


@pytest.fixture
def canonical_ensemble(composite_processor):
    return CanonicalEnsemble(composite_processor)


@pytest.fixture
def mugrand_ensemble(composite_processor):
    species = {
        sp for space in composite_processor.active_site_spaces for sp in space.keys()
    }
    chemical_potentials = {sp: 0.3 for sp in species}
    return SemiGrandEnsemble(
        composite_processor, chemical_potentials=chemical_potentials
    )


# Test with composite processors to cover more ground
@pytest.mark.parametrize("ensemble_cls", ensembles)
def test_from_cluster_expansion(cluster_subspace_ewald, ensemble_cls, rng):
    coefs = rng.random(cluster_subspace_ewald.num_corr_functions + 1)
    scmatrix = 3 * np.eye(3)
    proc = CompositeProcessor(cluster_subspace_ewald, scmatrix)
    proc.add_processor(
        ClusterExpansionProcessor(
            cluster_subspace_ewald, scmatrix, coefficients=coefs[:-1]
        )
    )
    proc.add_processor(
        EwaldProcessor(
            cluster_subspace_ewald,
            scmatrix,
            cluster_subspace_ewald.external_terms[0],
            coefficient=coefs[-1],
        )
    )
    reg_data = RegressionData(
        module="fake.module",
        estimator_name="Estimator",
        feature_matrix=rng.random((5, len(coefs))),
        property_vector=rng.random(len(coefs)),
        parameters={"foo": "bar"},
    )
    expansion = ClusterExpansion(cluster_subspace_ewald, coefs, reg_data)

    if ensemble_cls is SemiGrandEnsemble:
        species = {sp for space in proc.active_site_spaces for sp in space.keys()}
        chemical_potentials = {sp: 0.3 for sp in species}
        kwargs = {"chemical_potentials": chemical_potentials}
    else:
        kwargs = {}
    ensemble = ensemble_cls.from_cluster_expansion(
        expansion, supercell_matrix=scmatrix, **kwargs
    )
    npt.assert_array_equal(
        ensemble.natural_parameters[: ensemble.num_energy_coefs], coefs
    )
    occu = np.zeros(ensemble.num_sites, dtype=int)
    for _ in range(50):  # test a few flips
        sublatt = rng.choice(ensemble.active_sublattices)
        site = rng.choice(sublatt.sites)
        spec = rng.choice(sublatt.encoding)
        flip = [(site, spec)]
        assert proc.compute_property_change(
            occu, flip
        ) == ensemble.processor.compute_property_change(occu, flip)
        assert proc.compute_property(occu) == ensemble.processor.compute_property(occu)
        occu[site] = spec


def test_restrict_sites(ensemble, rng):
    sites = rng.choice(range(ensemble.processor.num_sites), size=5)
    ensemble.restrict_sites(sites)
    for sublatt in ensemble.sublattices:
        assert not any(i in sublatt.active_sites for i in sites)
    ensemble.reset_restricted_sites()
    for sublatt in ensemble.sublattices:
        if len(sublatt.site_space) > 1:
            npt.assert_array_equal(sublatt.sites, sublatt.active_sites)
        else:
            assert len(sublatt.active_sites) == 0


def test_msonable(ensemble):
    assert_msonable(ensemble)


def test_split_ensemble(ensemble):
    occu = gen_random_occupancy(ensemble.sublattices)
    for sublattice in ensemble.sublattices:
        npt.assert_array_equal(np.arange(len(sublattice.species)), sublattice.encoding)
        # ensemble must have been initialized from default.
    while len(ensemble.active_sublattices) > 0:
        is_active = [s.is_active for s in ensemble.sublattices]
        sl_id = np.random.choice(np.arange(len(is_active), dtype=int)[is_active])
        sublattice = ensemble.sublattices[sl_id]
        S = len(sublattice.species)
        old_sublattices = deepcopy(ensemble.sublattices)
        old_species = deepcopy(ensemble.species)
        if isinstance(ensemble, SemiGrandEnsemble):
            old_chemical_potentials = deepcopy(ensemble.chemical_potentials)
            old_mu_table = deepcopy(ensemble._mu_table)
        split_encodings = [sublattice.encoding[: S // 2], sublattice.encoding[S // 2 :]]
        ensemble.split_sublattice_by_species(sl_id, occu, split_encodings)
        assert len(ensemble.sublattices) == len(old_sublattices) + 1
        for i in range(len(old_sublattices)):
            if i != sl_id:
                if i < sl_id:
                    new_sublattice = ensemble.sublattices[i]
                    old_sublattice = old_sublattices[i]
                else:
                    new_sublattice = ensemble.sublattices[i + 1]
                    old_sublattice = old_sublattices[i]
                assert new_sublattice.site_space == old_sublattice.site_space
                npt.assert_array_equal(new_sublattice.sites, old_sublattice.sites)
                npt.assert_array_equal(
                    new_sublattice.active_sites, old_sublattice.active_sites
                )
                npt.assert_array_equal(new_sublattice.encoding, old_sublattice.encoding)
            else:
                old_sublattice = old_sublattices[i]
                new1 = ensemble.sublattices[i]
                new2 = ensemble.sublattices[i + 1]
                npt.assert_array_equal(
                    np.sort(old_sublattice.sites),
                    np.sort(np.concatenate((new1.sites, new2.sites))),
                )
                if new1.is_active and new2.is_active:
                    npt.assert_array_equal(
                        np.sort(old_sublattice.active_sites),
                        np.sort(np.concatenate((new1.active_sites, new2.active_sites))),
                    )
                npt.assert_array_equal(
                    np.sort(old_sublattice.encoding),
                    np.sort(np.concatenate((new1.encoding, new2.encoding))),
                )
                assert len(new1.encoding) == len(new1.species)
                assert len(new2.encoding) == len(new2.species)
        if isinstance(ensemble, SemiGrandEnsemble):
            assert set(ensemble.chemical_potentials.keys()) == set(ensemble.species)
            assert ensemble._mu_table.shape == old_mu_table.shape
            for sp in ensemble.species:
                assert sp in old_species
                assert ensemble.chemical_potentials[sp] == old_chemical_potentials[sp]
                for sublattice in ensemble.active_sublattices:
                    if sp in sublattice.species:
                        code = sublattice.encoding[sublattice.species.index(sp)]
                        npt.assert_array_equal(
                            ensemble._mu_table[sublattice.sites, code],
                            ensemble.chemical_potentials[sp],
                        )
            for sp in set(old_species) - set(ensemble.species):
                for sublattice in old_sublattices:
                    if sp in sublattice.species:
                        code = sublattice.encoding[sublattice.species.index(sp)]
                        npt.assert_array_equal(
                            ensemble._mu_table[sublattice.sites, code], 0
                        )
            for sublattice in ensemble.sublattices:
                if not sublattice.is_active:
                    npt.assert_array_equal(ensemble._mu_table[sublattice.sites, :], 0)


# Canonical Ensemble tests
def test_compute_feature_vector_canonical(canonical_ensemble, rng):
    processor = canonical_ensemble.processor
    occu = gen_random_occupancy(canonical_ensemble.sublattices)
    assert np.dot(
        canonical_ensemble.natural_parameters,
        canonical_ensemble.compute_feature_vector(occu),
    ) == pytest.approx(processor.compute_property(occu))
    npt.assert_array_equal(
        canonical_ensemble.compute_feature_vector(occu),
        canonical_ensemble.processor.compute_feature_vector(occu),
    )
    for _ in range(50):  # test a few flips
        sublatt = np.random.choice(canonical_ensemble.active_sublattices)
        site = rng.choice(sublatt.sites)
        spec = rng.choice(range(len(sublatt.site_space)))
        flip = [(site, spec)]
        assert np.dot(
            canonical_ensemble.natural_parameters,
            canonical_ensemble.compute_feature_vector_change(occu, flip),
        ) == pytest.approx(processor.compute_property_change(occu, flip))
        npt.assert_array_equal(
            canonical_ensemble.compute_feature_vector_change(occu, flip),
            processor.compute_feature_vector_change(occu, flip),
        )

    # Can still work normally with processor after splitting.
    is_active = [s.is_active for s in canonical_ensemble.sublattices]
    sl_id = np.random.choice(
        np.arange(len(canonical_ensemble.sublattices), dtype=int)[is_active]
    )
    encoding = canonical_ensemble.sublattices[sl_id].encoding
    S = len(encoding)
    split = [encoding[: S // 2], encoding[S // 2 :]]
    canonical_ensemble.split_sublattice_by_species(sl_id, occu, split)
    if len(canonical_ensemble.active_sublattices) > 0:
        assert np.dot(
            canonical_ensemble.natural_parameters,
            canonical_ensemble.compute_feature_vector(occu),
        ) == pytest.approx(processor.compute_property(occu))
        npt.assert_array_equal(
            canonical_ensemble.compute_feature_vector(occu),
            canonical_ensemble.processor.compute_feature_vector(occu),
        )
        for _ in range(50):  # test a few flips
            sublatt = np.random.choice(canonical_ensemble.active_sublattices)
            site = np.random.choice(sublatt.sites)
            spec = np.random.choice(range(len(sublatt.site_space)))
            flip = [(site, spec)]
            assert np.dot(
                canonical_ensemble.natural_parameters,
                canonical_ensemble.compute_feature_vector_change(occu, flip),
            ) == pytest.approx(processor.compute_property_change(occu, flip))
            npt.assert_array_equal(
                canonical_ensemble.compute_feature_vector_change(occu, flip),
                processor.compute_feature_vector_change(occu, flip),
            )


# MuSemigrandEnsemble Tests
def test_compute_feature_vector_sgc(mugrand_ensemble, rng):
    proc = mugrand_ensemble.processor
    occu = gen_random_occupancy(mugrand_ensemble.sublattices)
    assert np.dot(
        mugrand_ensemble.natural_parameters,
        mugrand_ensemble.compute_feature_vector(occu),
    ) == pytest.approx(
        proc.compute_property(occu) - mugrand_ensemble.compute_chemical_work(occu)
    )
    npt.assert_array_equal(
        mugrand_ensemble.compute_feature_vector(occu)[:-1],
        mugrand_ensemble.processor.compute_feature_vector(occu),
    )
    for _ in range(50):  # test a few flips
        sublatt = rng.choice(mugrand_ensemble.active_sublattices)
        site = rng.choice(sublatt.sites)
        spec = rng.choice(sublatt.encoding)
        flip = [(site, spec)]
        dmu = (
            mugrand_ensemble._mu_table[site][spec]
            - mugrand_ensemble._mu_table[site][occu[site]]
        )
        assert np.dot(
            mugrand_ensemble.natural_parameters,
            mugrand_ensemble.compute_feature_vector_change(occu, flip),
        ) == pytest.approx(proc.compute_property_change(occu, flip) - dmu)
        npt.assert_array_equal(
            mugrand_ensemble.compute_feature_vector_change(occu, flip),
            np.append(proc.compute_feature_vector_change(occu, flip), dmu),
        )
    # Can still work normally with processor after splitting.
    is_active = [s.is_active for s in mugrand_ensemble.sublattices]
    sl_id = np.random.choice(
        np.arange(len(mugrand_ensemble.sublattices), dtype=int)[is_active]
    )
    encoding = mugrand_ensemble.sublattices[sl_id].encoding
    S = len(encoding)
    split = [encoding[: S // 2], encoding[S // 2 :]]
    mugrand_ensemble.split_sublattice_by_species(sl_id, occu, split)
    if len(mugrand_ensemble.active_sublattices) > 0:
        assert np.dot(
            mugrand_ensemble.natural_parameters,
            mugrand_ensemble.compute_feature_vector(occu),
        ) == pytest.approx(
            proc.compute_property(occu) - mugrand_ensemble.compute_chemical_work(occu)
        )
        npt.assert_array_equal(
            mugrand_ensemble.compute_feature_vector(occu)[:-1],
            mugrand_ensemble.processor.compute_feature_vector(occu),
        )
        for _ in range(50):  # test a few flips
            sublatt = np.random.choice(mugrand_ensemble.active_sublattices)
            site = np.random.choice(sublatt.sites)
            spec = np.random.choice(sublatt.encoding)
            flip = [(site, spec)]
            dmu = (
                mugrand_ensemble._mu_table[site][spec]
                - mugrand_ensemble._mu_table[site][occu[site]]
            )
            assert np.dot(
                mugrand_ensemble.natural_parameters,
                mugrand_ensemble.compute_feature_vector_change(occu, flip),
            ) == pytest.approx(proc.compute_property_change(occu, flip) - dmu)
            npt.assert_array_equal(
                mugrand_ensemble.compute_feature_vector_change(occu, flip),
                np.append(proc.compute_feature_vector_change(occu, flip), dmu),
            )


def test_bad_chemical_potentials(mugrand_ensemble):
    proc = mugrand_ensemble.processor
    chem_pots = mugrand_ensemble.chemical_potentials
    print(chem_pots)
    with pytest.raises(ValueError):
        items = list(chem_pots.items())
        mugrand_ensemble.chemical_potentials = {items[0][0]: items[0][1]}
    with pytest.raises(ValueError):
        mugrand_ensemble.chemical_potentials = {"A": 0.5, "D": 0.6}
    with pytest.raises(ValueError):
        chem_pots["foo"] = 0.4
        SemiGrandEnsemble(proc, chemical_potentials=chem_pots)
    with pytest.raises(ValueError):
        del chem_pots["foo"]
        chem_pots.pop(items[0][0])
        SemiGrandEnsemble(proc, chemical_potentials=chem_pots)
    with pytest.raises(ValueError):
        chem_pots[str(list(chem_pots.keys()))[0]] = 0.0
        mugrand_ensemble.chemical_potentials = chem_pots


def test_build_mu_table(mugrand_ensemble):
    # Do not use processor sub-lattices in these tests.
    table = mugrand_ensemble._build_mu_table(mugrand_ensemble.chemical_potentials)
    for sublatt, row in zip(mugrand_ensemble.active_sublattices, table):
        if len(sublatt.species) == 1:  # skip inactive sites
            continue
        for i, species in zip(sublatt.encoding, sublatt.species):
            assert mugrand_ensemble.chemical_potentials[species] == row[i]



"""Test all bias terms."""

from copy import deepcopy

import pytest

from smol.moca.sampler.bias import FugacityBias, mcbias_factory
from tests.utils import gen_random_occupancy

bias_classes = [FugacityBias]


@pytest.fixture(scope="module")
def all_sublattices(ce_processor):
    return ce_processor.get_sublattices()


@pytest.fixture(params=bias_classes)
def mcbias(all_sublattices, request):
    return request.param(all_sublattices)


def test_compute_bias_change(mcbias, rng):
    step = []
    occu = gen_random_occupancy(mcbias.sublattices)
    new_occu = occu.copy()
    for _ in range(50):
        s = rng.choice(list(range(len(mcbias.active_sublattices))))
        i = rng.choice(mcbias.active_sublattices[s].sites)
        sp = rng.choice(list(range(len(mcbias.active_sublattices[s].species))))
        step.append((i, sp))
        if i == 81:
            raise (ValueError, "81!!!!")
        new_occu[i] = sp
    print(mcbias.sublattices)
    print(step)
    assert mcbias.compute_bias_change(occu, step) == pytest.approx(
        mcbias.compute_bias(new_occu) - mcbias.compute_bias(occu)
    )


def test_mcbias_factory(all_sublattices):
    for bias in bias_classes:
        assert isinstance(mcbias_factory(bias.__name__, all_sublattices), bias)


# Tests for FugacityBias
# Tests for FuSemiGrandEnsemble
@pytest.fixture(scope="module")
def fugacity_bias(all_sublattices):
    return FugacityBias(all_sublattices)


def test_bad_fugacity_fractions(fugacity_bias):
    fug_fracs = deepcopy(fugacity_bias.fugacity_fractions)
    with pytest.raises(ValueError):
        fug_fracs[0] = {s: v for s, v in list(fug_fracs[0].items())[:-1]}
        fugacity_bias.fugacity_fractions = fug_fracs
    with pytest.raises(ValueError):
        fug_fracs[0] = {sp: 1.1 for sp in fug_fracs[0].keys()}
        fugacity_bias.fugacity_fractions = fug_fracs
    with pytest.raises(ValueError):
        fug_fracs[0] = {"A": 0.5, "D": 0.6}
        fugacity_bias.fugacity_fractions = fug_fracs
    with pytest.raises(ValueError):
        fug_fracs[0]["foo"] = 0.4
        FugacityBias(
            fugacity_bias.sublattices,
            fugacity_fractions=fug_fracs,
        )
    with pytest.raises(ValueError):
        del fug_fracs[0]["foo"]
        FugacityBias(
            fugacity_bias.sublattices,
            fugacity_fractions=fug_fracs,
        )
    with pytest.raises(ValueError):
        fug_fracs[0][str(list(fug_fracs[0].keys())[0])] = 0.0
        fugacity_bias.fugacity_fractions = fug_fracs


def test_build_fu_table(fugacity_bias):
    table = fugacity_bias._build_fu_table(fugacity_bias.fugacity_fractions)
    for sublatt in fugacity_bias.active_sublattices:
        for fus in fugacity_bias.fugacity_fractions:
            if list(sublatt.site_space.keys()) == list(fus.keys()):
                fugacity_fractions = fus
        for i in sublatt.sites:
            for j, species in zip(sublatt.encoding, sublatt.site_space):
                assert fugacity_fractions[species] == table[i, j]



import numpy as np
import numpy.testing as npt
from pymatgen.analysis.ewald import EwaldSummation
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import PeriodicSite, Structure

from smol.cofe.extern import EwaldTerm
from smol.cofe.space.domain import Vacancy, get_allowed_species
from tests.utils import assert_msonable, gen_random_neutral_occupancy


def test_get_ewald_structure(ce_processor):
    supercell = ce_processor.structure
    sm = StructureMatcher()
    ew = EwaldTerm(eta=0.15)
    ew_structure, ew_inds = ew.get_ewald_structure(supercell)

    spaces = get_allowed_species(supercell)
    nbits = np.array([len(space) - 1 for space in spaces])
    sites = []
    for space, site in zip(spaces, supercell):
        for b in space:
            if isinstance(b, Vacancy):  # skip vacancies
                continue
            sites.append(PeriodicSite(b, site.frac_coords, supercell.lattice))

    s = Structure.from_sites(sites)
    assert sm.fit(ew_structure, s)

    assert ew_inds.shape == (len(supercell), max(nbits) + 1)
    start = 0
    for space, inds in zip(spaces, ew_inds):
        if isinstance(space[-1], Vacancy):
            n_sp = len(space) - 1
        else:
            n_sp = len(space)
        npt.assert_allclose(inds[:n_sp], np.arange(start, start + n_sp))
        npt.assert_allclose(inds[n_sp:], -1)
        start += n_sp


def test_val_from_occupancy(ce_processor):
    # Test 10 times at random, with charge balance.
    supercell = ce_processor.structure
    sublattices = ce_processor.get_sublattices()
    n_success = 0
    for _ in range(10):
        try:
            # We should only test neutral occupancies.
            occu = gen_random_neutral_occupancy(sublattices)
            n_success += 1
        except TimeoutError:
            occu = None

        if occu is not None:
            s = ce_processor.structure_from_occupancy(occu)
            assert s.charge == 0
            ew = EwaldTerm(eta=0.15)
            np.testing.assert_almost_equal(
                ew.value_from_occupancy(occu, supercell),
                EwaldSummation(s, eta=ew.eta).total_energy,
                decimal=7,
            )
            ew = EwaldTerm(eta=0.15, use_term="real")
            np.testing.assert_almost_equal(
                ew.value_from_occupancy(occu, supercell),
                EwaldSummation(s, eta=ew.eta).real_space_energy,
                decimal=7,
            )
            ew = EwaldTerm(eta=0.15, use_term="reciprocal")
            np.testing.assert_almost_equal(
                ew.value_from_occupancy(occu, supercell),
                EwaldSummation(s, eta=ew.eta).reciprocal_space_energy,
                decimal=7,
            )
            ew = EwaldTerm(eta=0.15, use_term="point")
            np.testing.assert_almost_equal(
                ew.value_from_occupancy(occu, supercell),
                EwaldSummation(s, eta=ew.eta).point_energy,
                decimal=7,
            )
    assert n_success > 0


def test_msonable():
    ew = EwaldTerm(eta=0.15, real_space_cut=0.5, use_term="point")
    assert_msonable(ew)



import os

from smol.cofe import ClusterSubspace, StructureWrangler
from smol.io import load_work, save_work
from smol.moca import CanonicalEnsemble, ClusterExpansionProcessor


def test_save_load_work(single_canonical_ensemble, tmpdir):
    processor = single_canonical_ensemble.processor
    subspace = processor.cluster_subspace
    wrangler = StructureWrangler(processor)
    file_path = os.path.join(tmpdir, "smol.mson")
    save_work(file_path, subspace, wrangler, processor, single_canonical_ensemble)
    assert os.path.isfile(file_path)
    work_dict = load_work(file_path)
    assert len(work_dict) == 4
    for name, obj in work_dict.items():
        assert name == obj.__class__.__name__
        assert type(obj) in (
            ClusterSubspace,
            StructureWrangler,
            ClusterExpansionProcessor,
            CanonicalEnsemble,
        )
    os.remove(file_path)



"""Classes to define subspaces of functions over crystalline configurations.

Contains classes and functions necessary to represent the configurational space
of a crystalline system and the space of functions over it.
"""

from .basis import basis_factory
from .cluster import Cluster
from .domain import Vacancy, get_allowed_species, get_site_spaces, get_species
from .orbit import Orbit

__all__ = [
    "Cluster",
    "Orbit",
    "Vacancy",
    "get_species",
    "get_site_spaces",
    "get_allowed_species",
    "basis_factory",
]
